/* mpeg2dec.c, main(), initialization, option processing                    */

/* Copyright (C) 1994, MPEG Software Simulation Group. All Rights Reserved. */

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 * The MPEG Software Simulation Group does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any third-party
 * patents.
 *
 * Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
 * are subject to royalty fees to patent holders.  Many of these patents are
 * general enough such that they are unavoidable regardless of implementation
 * design.
 *
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef ERROR

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>

#define GLOBAL
#include "config.h"
#include "global.h"
#include "ifrmdec.h"


static CRITICAL_SECTION CritSect;

/* private prototypes */
static void initdecoder(DECODE_FRAME *df);
static void ZeroGlobals(void);



void InitMpegDecode(void)
/*
init mpeg decoder global tables
*/
{
	int i;

	ZeroGlobals();
	BuildDCTTable();
	InitRGBTabels();
    init_idct();

	//init critical section
	InitializeCriticalSection(&CritSect);

	//init clip tables
	clp = ClipTable + 384;
	ClpRGB16 = ClipTableRGB16 + 384;
	for (i=-384; i<640; i++) {
		clp[i] = (i<0) ? 0 : ((i>255) ? 255 : i);
#if 0  // this is code to force dithering
		if ((i & 0x7)==7 )
			ClpRGB16[i] = (i+8<0) ? 0 : ((i+8>255) ? (255>>3) : ((i + 8)>>3));
		else if ((i & 0x7)==0)
			ClpRGB16[i] = (i-8<0) ? 0 : ((i-8>255) ? (255>>3) : ((i -8)>>3));
		else
#endif
			ClpRGB16[i] = (i<0) ? 0 : ((i>255) ? (255>>3) : (i>>3));
	}


}

void FreeMpegDecode(void) {
	DeleteCriticalSection(&CritSect);
}


DECODE_FRAME *OpenDecoder(void)
{
	DECODE_FRAME *df = (DECODE_FRAME *)malloc(sizeof(DECODE_FRAME));
	if (df == NULL)
		return NULL;

	
	df->FrameBuff[0] = 0;
	df->FrameBuff[1] = 0;
	df->FrameBuff[2] = 0;
	df->FrameWidth = 0;
	df->FrameHeight = 0;
	return df;
}

void CloseDecoder(DECODE_FRAME *df)
/*
Free frame memory 
*/
{
	free(df->FrameBuff[0]);
	free(df->FrameBuff[1]);
	free(df->FrameBuff[2]);
	free(df);

}

int SetFrameSize(DECODE_FRAME *df, int width, int height)
/*
Set size of MPEG frame.  Allocates / reallocates frame memory and
sets frame size parameters.  Returns 1 if successfull, 0 otherwise.
*/

{
	int cc, size;


	//return if frame same hasn't changed
	if (df->FrameWidth == width && df->FrameHeight == height)
		return 1;
	
	
	//init globals to current size
	df->FrameWidth = width;
	df->FrameHeight = height;
	initdecoder(df);

	//allocate frame buffers
	for (cc=0; cc<3; cc++) {
		if (cc==0)
			size = coded_picture_width*coded_picture_height;
		else
			size = chrom_width*chrom_height;

		if (!(df->FrameBuff[cc] = (unsigned char *)realloc(df->FrameBuff[cc], size)))	{
			//if we can't alloc frame, clean up and return error
			free(df->FrameBuff[0]);
			free(df->FrameBuff[1]);
			free(df->FrameBuff[2]);
			df->FrameBuff[0] = 0;
			df->FrameBuff[1] = 0;
			df->FrameBuff[2] = 0;
			df->FrameWidth = 0;
			df->FrameHeight = 0;
			return 0;
		}
	}

	return 1;
}

int DecodeFrame(DECODE_FRAME *df, unsigned char *FrameData, int FrameBytes,
							void *OutData, ESWDecodeOutputType OutputType)
/*
This function decodes the specified MPEG Iframe.  It writes the decoded data to the specified output
buffer using a format specified by the OutPelBits, which specifies the number of bits
per pixel in the output image.  The function returns TRUE (1) if successful,
FALSE (0) oterwise.
*/

{
	EnterCriticalSection(&CritSect);

	//init globals to current frame params
	switch (OutputType) {
	case RGB555:
		Output16Frame = (short *)OutData;
		CovertToRGB = CovertBlockToRGB16;  //set output function to RGB16
		break;
	case RGB565:
		Output16Frame = (short *)OutData;
		CovertToRGB = CovertBlockToRGBBF16;  //set output function to RGB16 bit field
		break;
	case RGB24BIT:
		RGB24Frame = (RGB24_PIXEL *)OutData;
		CovertToRGB = CovertBlockToRGB24;  //set output function to RGB24
		break;
	case UYVY:
		Output16Frame = (short *)OutData;
		CovertToRGB = CovertBlockToUYVY;  //set output function to UYVY (422)
		break;
	default:
		LeaveCriticalSection(&CritSect);
		return 0;
	}

	initdecoder(df);
	ld = &base;
	InitBitsToFrame(FrameData, FrameBytes);

	//decode the frame

	__try {
		getheader();
		if (horizontal_size!=df->FrameWidth || vertical_size!=df->FrameHeight) {
			LeaveCriticalSection(&CritSect);
			return 0;	 //error if mpeg stream params don't match image params
		}

		getIFrameMBs(FrameNum++);
		LeaveCriticalSection(&CritSect);
		return 1;
	}
	__except (  GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION
			  ||GetExceptionCode() == EXCEPTION_ARRAY_BOUNDS_EXCEEDED) {
		//premature frame end or bad VLC data can cause an access violation.
		//in any case, catch the exception and return failed
		LeaveCriticalSection(&CritSect);
		return 0;
	}

}

static void initdecoder(DECODE_FRAME *df)
{
  static int blk_cnt_tab[3] = {6,8,12};

  /* force MPEG-1 parameters */
  base.mpeg2 = 0;
  if (!base.mpeg2)
  {
    prog_seq = 1;
    prog_frame = 1;
    pict_struct = FRAME_PICTURE;
    frame_pred_dct = 1;
    chroma_format = CHROMA420;
    matrix_coefficients = 5;
  }

  /* round to nearest multiple of coded macroblocks */
  mb_width = (df->FrameWidth+15)/16;
  mb_height = (base.mpeg2 && !prog_seq) ? 2*((df->FrameHeight+31)/32)
                                        : (df->FrameHeight+15)/16;
  coded_picture_width = 16*mb_width;
  coded_picture_height = 16*mb_height;

  chrom_width = (chroma_format==CHROMA444) ? coded_picture_width
                                           : coded_picture_width>>1;
  chrom_height = (chroma_format!=CHROMA420) ? coded_picture_height
                                            : coded_picture_height>>1;
  blk_cnt = blk_cnt_tab[chroma_format-1];

  newframe[0] = df->FrameBuff[0];
  newframe[1] = df->FrameBuff[1];
  newframe[2] = df->FrameBuff[2];

}

void error(text)
char *text;
{
  //fprintf(stderr,text);
  //exit(1);
}

/* trace output */
void printbits(code,bits,len)
int code,bits,len;
{
  int i;
  for (i=0; i<len; i++)
    printf("%d",(code>>(bits-1-i))&1);
}



static void ZeroGlobals(void)
/*
This code depends on many of the globlas beign initialy zero.  So
let's do the brute force thing and zero them all.
*/
{
	newframe[0] = 0;
	newframe[1] = 0;
	newframe[2] = 0;
	FrameNum = 0;
	quiet = 0;
	trace = 0;
	memset(errortext, 0, sizeof(errortext));
	clp = 0;
	profile = 0;
	level = 0;
	horizontal_size = 0;
	vertical_size = 0;
	mb_width = 0;
	mb_height = 0;
	coded_picture_width = 0;
	coded_picture_height = 0;
	chroma_format = 0;
	chrom_width = 0;
	chrom_height = 0;
	blk_cnt = 0;

	pict_type = 0;
	forw_r_size = 0;
	back_r_size = 0;
	full_forw = 0;
	full_back = 0;
	fault = 0;
	verbose = 0;
	prog_seq = 0;

	h_forw_r_size = 0;
	v_forw_r_size = 0;
	h_back_r_size = 0;
	v_back_r_size = 0;
	dc_prec = 0;
	pict_struct = 0;
	topfirst = 0;
	frame_pred_dct = 0;
	conceal_mv = 0;
	intravlc = 0;
	repeatfirst = 0;
	prog_frame = 0;
	secondfield = 0;
	stwc_table_index = 0;
	llw = 0;
	llh = 0;
	hm = 0;
	hn = 0;
	vm = 0;
	vn = 0;

	lltempref = 0;
	llx0 = 0;
	lly0 = 0;
	llprog_frame = 0;
	llfieldsel = 0;
	llinputname = 0;

	twostreams = 0;
	sflag = 0;
	refidct = 0;
	matrix_coefficients = 0;

	/* output */
	framestoreflag = 0;
	outputname = 0;
	outtype = 0;
	hiQdither = 0;

	/* layer specific variables (needed for SNR and DP scalability) */
	memset(&base, 0, sizeof(base));
	memset(&enhan, 0, sizeof(enhan));
	ld = 0;
}


