#ifndef __SMR__
#define __SMR__

/* smr.h
This file contaions the interface to the CCalcSMR class.  This class performs the 
psychoacoustic model 1 (for layer2 only) form ISO/IEC 11172-3 Annex D.
*/

#include "lyr2defs.h"
#include <math.h>

#define FFTSIZE 1024
#define LOG2FFTSIZE 10
#define MAXTONES 55   //max 1 tone per .5 bark (highest bark value = 24.5)
#define MAXNONTONES 27 //see 11172-3 table D.2
#define MAXSUBSAMPSIZE 132 //see 11172-3 table D.1
#define BINSPERSUBBAND ((FFTSIZE/2)/SUBBANDS)

/*
compute the signal to mask ratio (SMR) for each subband from the specified raw input
(already delayed to match the sbband filter) with the specifed scale factors.  The 
results are written to the specified SMR array.
*/
class CCalcSMR {
	private:
    //----sub-sampled frequency tables, for each sample freq  (table D1 form 11172-3)
    //(note that these tables are 0 based, not 1 based as in the spec)
    static const struct SSubSampleTable {   // 
        int fftBin;
        float CritBandRate;         //in bark
        float AbsThrs;              //absolute threshold in dB
    } SubSamp32000[132], SubSamp44100[130], SubSamp48000[126];

	static  float AbsThrsLin[132]; // must be the largest of the tables above

    //critical band boundries tables for each sample freq (table D2 form 11172-3)
	static const struct SCritBandTable {
		int		MidBin;			//bin at center of critical band
		int		MaxBin;			//bit at top of critical band
	} CritBands32000[25], CritBands44100[27], CritBands48000[27];

    static const float dBScaleFactorTable[63];  //MPEG-1 layer II scale factors in RMS dB
                                                //(see 11172-3 annex D step 2.

    int BinToIndex[FFTSIZE/2]; //index into subsampled table for each FFT bin

   
    //----data members---------------------
    float HannWindow[FFTSIZE]; //holds the hann window function
    float fftPow[FFTSIZE];     //fft output in form of power (also used for temp input storage)
    float fftdB[FFTSIZE/2];    //fft output in dB
    float AbsThreshOffset;     //offset from absolute threshold in table D1
    const SSubSampleTable *SubSamp; //pointer to current sub-sampled freq table
    int SubSampSize;           //size of current sub-sampled freq table
    const SCritBandTable *CritBands;      //pointer to current critical band table
    int NumCritBands;          //number of bands in current critical band table

    struct SSignalComponents {
        int SubSampIndex;       //sub-sampled index of tone (index into table D.1)
        float CritBandRate;     //critical band rate (from 11172-3 table D.1)
        float MagdB;            //component sound preasure level in db
		float MagdBLin;			// Linear version of the same
    } Tones[MAXTONES], NonTones[MAXNONTONES]; 
    int NumTones;
    int NumNonTones;

    float GlobalMaskThold[MAXSUBSAMPSIZE];  //storage for the output of the global masking 
                                            //threshold calculation (power - not in dB)
    float SubbandPower[SUBBANDS];        //signal power for each subband

	// Logtable used for fast dBToPower 
#define MINLOG  (-300.0F)
#define LOGTABOFZERO 300   // lobtab index of zero db
#define LOGTABSIZE (1000)

	struct logtabrcd {
		float value;
		float slope;
	} logtab[LOGTABSIZE], *logtabzero;
 



    //----interface functions-------
    public:
	CCalcSMR(ESamplingFreq Freq = SF_44100_HZ, int BitRate = 160 /*(in KHz)*/);
    
    void CalcSignalToMask(float RawData[INPUTSIZE],
							int ScaleFactors[SUBBANDS][NUMGROUPS],
							float SMR[SUBBANDS],  int ValidSubbands);

    //----Helper Functions
    int FindTones(SSignalComponents component[MAXTONES]);
    int IsTone(int bin, int ToneWidth);
    int FindNonTones(SSignalComponents component[MAXNONTONES]);
	int BinToWidth(int bin);
	void StoreToneData(SSignalComponents *component, int bin);
	void UpdateGlobalMasking(SSignalComponents &comp, int IsTone);
    void CalcSubbandPower(int ScaleFactors[SUBBANDS][NUMGROUPS], int ValidSubbands);
	//----Fast dBToPower conversion

    inline float dBToPower(float x) {
	int k = (int)(x);
	if (k <= -(LOGTABOFZERO))
		return(logtab[0].value);
	else
		return (logtabzero[k].value + (x - (float)(k)) * logtabzero[k].slope);
	}	


    
};
      

#endif __SMR__