/*ConvAud.cpp
This file contains code to convert audio formats
*/
#define WIN32_LEAN_AND_MEAN
#define DOSWIN32
#include <windows.h>
#include <vfw.h>
#include <math.h>
#include "convaud.h"

#define PI 3.141592654


#define BLOCK_SIZE 1024

BOOL ConvertAudio(PAVISTREAM InStream, PAVISTREAM OutStream)
/*
This function copies all the audio data from the specified input stream to
the specified output stream.  The function converts the audio to match the
format of the output stream.  If successfull, the function returns TRUE, FALSE
otherwise.
*/
{
    //init converet sample access objects
    CAccessSamples InputSamples;
    CAccessSamples OutputSamples;

    if (!InputSamples.init(InStream))
        return FALSE;
    if (!OutputSamples.init(OutStream))
        return FALSE;
        
    CConvertRate RateConvert(InputSamples.GetSampleRate(), OutputSamples.GetSampleRate()); 

    //alloc sample data
    SAMPLE *InBuff = new SAMPLE[BLOCK_SIZE];
    if (InBuff == NULL)
        return FALSE;
    SAMPLE *OutBuff = new SAMPLE[RateConvert.GetMaxOutSize(BLOCK_SIZE)];
    if (OutBuff == NULL) {
        delete [] InBuff;
        return FALSE;
    }

	//convert the data
    long CurInSamp = AVIStreamStart(InStream);
    long CurOutSamp = 0;
    long SamplesRead = InputSamples.read(InBuff, CurInSamp, BLOCK_SIZE);
    while (SamplesRead > 0) {
        //convert channel 0 sample rate
        long OutSamples = RateConvert.Convert(InBuff, OutBuff, 0, SamplesRead);

        if (InputSamples.IsMono()) {
            //if input mono, save time by copying channel 1 = channel 0
            for (long i=0; i<OutSamples; i++)
                OutBuff[i][1] = OutBuff[i][0];
        }
        else {
            //otherwise convert channel 1 sample rate
            RateConvert.Convert(InBuff, OutBuff, 1, SamplesRead);
        }

        //write data to output stream
        if (OutputSamples.write(OutBuff, CurOutSamp, OutSamples) != OutSamples)
            break;   //break loop if write error

        //update loop counters
        CurInSamp += SamplesRead;
        CurOutSamp += OutSamples;

        //read next block of input data
        SamplesRead = InputSamples.read(InBuff, CurInSamp, BLOCK_SIZE);
    }

	//clean up
	delete [] InBuff;
	delete [] OutBuff;

	return (SamplesRead == 0);
}



CConvertRate::CConvertRate(double InRate, double OutRate)
//constructor, init member data
{
	m_OutputScale = OutRate / InRate;
	m_InputScale = InRate / OutRate;

	if (m_OutputScale == 2.0) {
		InitFilter(.1F, 0.0F, 2.0F);
	}
	if (m_OutputScale == 4.0) {
		InitFilter(.05F, 0.0F, 4.0F);
	}

	m_InterpolSave[0] = 0.0F;
	m_InterpolSave[1] = 0.0F;
	m_CurInSample = 0.0;

	m_FiltBuff = NULL;
	m_FiltBuffSize = 0;

}

CConvertRate::~CConvertRate()
{
	if (m_FiltBuff)
		free(m_FiltBuff);
}

long CConvertRate::Convert(SAMPLE *InBuff, SAMPLE *OutBuff, int channel, long InSize)
/*
convert the sample rate on the specified input samples.  Returns number of output samples
*/
{
//	double OutputScale = m_OutputScale;
	double OutputScale = 0.0;	   //set to always use intterpolation

	long OutSize;


	//calc required filter data size
	if (OutputScale == 2.0)
		OutSize = 2*InSize;
	else if (OutputScale == 4.0) 
		OutSize = 4*InSize;
	else 
		OutSize = 0;

	//reallocat buffer if needed
	if (OutSize > m_FiltBuffSize) {
		m_FiltBuff = (SAMPLE *)realloc(m_FiltBuff, OutSize * sizeof(SAMPLE));
		if (!m_FiltBuff) {
			m_FiltBuffSize = 0;
			return 0;
		}
		m_FiltBuffSize = OutSize;
	}

	if (OutputScale == 2.0) {
		//double sample rate - fill with zeros and filter
		for (int i=0; i<InSize; i++) {
			m_FiltBuff[2*i][channel] = InBuff[i][channel];
			m_FiltBuff[2*i+1][channel] = 0.0F;
		}
		filter(m_FiltBuff, OutBuff, channel, OutSize);
	}
	else if (OutputScale == 4.0) {
		//4 x sample rate - fill with zeros and filter
		for (int i=0; i<InSize; i++) {
			m_FiltBuff[4*i][channel] = InBuff[i][channel];
			m_FiltBuff[4*i+1][channel] = 0.0F;
			m_FiltBuff[4*i+2][channel] = 0.0F;
			m_FiltBuff[4*i+3][channel] = 0.0F;
		}
		filter(m_FiltBuff, OutBuff, channel, OutSize);
	}
    else if (OutputScale == 1.0) {
		for (int i=0; i<InSize; i++) {
			OutBuff[i][channel] = InBuff[i][channel];
		}
        OutSize = InSize;
    }
	else {
		//all other conversions - use linear interpolation
		while (m_CurInSample < 0) {
			float factor = 1 + (float)m_CurInSample;
			OutBuff[OutSize++][channel] = (1.0F - factor) * m_InterpolSave[channel]
										+ factor * InBuff[0][channel];
			m_CurInSample += m_InputScale;
		}
		while (m_CurInSample < InSize-1) {
			int InSamp = (int)m_CurInSample;
			float factor = (float)m_CurInSample - InSamp;
			OutBuff[OutSize++][channel] = (1.0F - factor) * InBuff[InSamp][channel]
										+ factor * InBuff[InSamp+1][channel];
			m_CurInSample += m_InputScale;
		}
		//setup for next block
		m_InterpolSave[channel] = InBuff[InSize-1][channel];
		m_CurInSample -= InSize;
	}

	return OutSize;
}


/*define -74 dB Blackman Harris terms:*/
#if 1
#define A0 0.40217
#define A1 0.49703
#define A2 0.09892
#define A3 0.00188
#else
/*define -92 dB Blackman Harris terms:*/
#define A0 0.35875
#define A1 0.48829
#define A2 0.14128
#define A3 0.01168
#endif


void CConvertRate::InitFilter(float cutoff, float CenterFreq, float FilterGain)
/*
This function generates the filter object for a FIR filter with the
specified parms:
cutoff is the freq difference from the cutoff point to the filter center in normalized freq
CenterFreq is the center of the filter

The funciton fills the FIR coefs in the m_coefs member array.  The filter size is
set to FILTER_LENGTH.

The function generates a windowed filter using the
-74 dB/-92 dB blackman harris window.
*/
{
    int i;
    float gain;
    float m,pi2,wfreq,wphase,sfreq,sphase,cfreq,cphase;
    float CenterFreqTerm;
    
    /*compute frequencies and starting phases*/
    pi2 = (float)(2.0*PI);
    wfreq = pi2/FILTER_LENGTH;                                  /*window freq*/
    wphase = (float).5 * wfreq;                             /*window starting phase*/
    sfreq = pi2*cutoff;                                     /*sin(x)/x freq*/
    sphase = ((float)0.5 - (float)FILTER_LENGTH/2) * sfreq;     /*sin(x)/x starting phase*/
    m = (float)0.5 - (float)FILTER_LENGTH/2;
    cfreq = pi2*CenterFreq;                                 /*cos freq*/
    cphase = ((float)0.5 - (float)FILTER_LENGTH/2) * cfreq;     /*cos starting phase*/
    gain = (float)0.0;    /*init running sum*/

    /*calculate filter coefs*/
    for (i=0; i<FILTER_LENGTH ;i++) {
        /* generate window */
        m_coefs[i] = (float)(A0 - A1*cos(wphase) + A2*cos(2.0*wphase)
                           - A3*cos(3.0*wphase));
        wphase += wfreq;

        /* scale with sin(x)/x */
        if (m != 0)
            m_coefs[i] *=  (float)sin(sphase)/m;  /* scale with sin(sfreq*m)/m*/
        else
            m_coefs[i] *= sfreq;
        sphase += sfreq;
        if (sphase > pi2)
            sphase -= pi2;
        m += (float)1.0;
            
        /* shift with cos(CenterFreq*x)*/
        CenterFreqTerm = (float)cos(cphase);
        m_coefs[i] *= CenterFreqTerm;
        cphase += cfreq;
        if (cphase > pi2)
            cphase -= pi2;
            
        /*update running sum at centerfreq*/
        gain += CenterFreqTerm*m_coefs[i];

    }
    
    /*normalize the filter*/
    if (gain > 0)
        gain = FilterGain/gain;
    for (i=0; i<FILTER_LENGTH; i++)
        m_coefs[i] *= gain;
        
    /*init the filter buffers and pointers*/
    for (i = 0; i<=FILTER_LENGTH; i++) {
		m_FilterData[i][0] = 0.0F;
		m_FilterData[i][1] = 0.0F;
	}
}

BOOL CConvertRate::filter(SAMPLE *InBuff, SAMPLE *OutBuff, int channel, long InputSize)
/*
This funciton filters the specifed channel of the input buffer and writes results to the
specified output.  The input must be at least FILTER_LENGTH-1 points long.
NOTE:  This filter can not be run in-place 
*/
{
	//exit if input too short
	if (InputSize < FILTER_LENGTH-1)
		return FALSE; 

	long samp;

	//process the first FILTER_LENGTH-1 points using saved data
	for (samp = 0; samp<FILTER_LENGTH-1; samp++) {
		int i,j,k;
		float out = 0.0F;
		for (i = samp, j=0; i<FILTER_LENGTH-1; i++, j++)
			out += m_coefs[j] * m_FilterData[i][channel];
		for (k=0; j<FILTER_LENGTH; j++, k++)
			out += m_coefs[j] * InBuff[k][channel];
		OutBuff[samp][channel] = out;
	}

	//save the last FILTER_LENGTH-1 for the next block
	long i,j;
	for (j=0, i=InputSize - FILTER_LENGTH + 1; i < InputSize; j++, i++)
		m_FilterData[j][channel] = InBuff[i][channel];
		
	//process the bulk of the data (un-wrap for speed)
	for (long InSamp = 0; samp < InputSize; samp++, InSamp++) {
		float *DataPtr = &InBuff[InSamp][channel];
		float out = m_coefs[0] * DataPtr[0*2];
		out += m_coefs[ 1] * DataPtr[ 1*2];
		out += m_coefs[ 2] * DataPtr[ 2*2];
		out += m_coefs[ 3] * DataPtr[ 3*2];
		out += m_coefs[ 4] * DataPtr[ 4*2];
		out += m_coefs[ 5] * DataPtr[ 5*2];
		out += m_coefs[ 6] * DataPtr[ 6*2];
		out += m_coefs[ 7] * DataPtr[ 7*2];
		out += m_coefs[ 8] * DataPtr[ 8*2];
		out += m_coefs[ 9] * DataPtr[ 9*2];
		out += m_coefs[ 0] * DataPtr[ 0*2];
		out += m_coefs[11] * DataPtr[11*2];
		out += m_coefs[12] * DataPtr[12*2];
		out += m_coefs[13] * DataPtr[13*2];
		out += m_coefs[14] * DataPtr[14*2];
		out += m_coefs[15] * DataPtr[15*2];
		out += m_coefs[16] * DataPtr[16*2];
		out += m_coefs[17] * DataPtr[17*2];
		out += m_coefs[18] * DataPtr[18*2];
		out += m_coefs[19] * DataPtr[19*2];
		out += m_coefs[20] * DataPtr[20*2];
		out += m_coefs[21] * DataPtr[21*2];
		out += m_coefs[22] * DataPtr[22*2];
		out += m_coefs[23] * DataPtr[23*2];
		out += m_coefs[24] * DataPtr[24*2];
		out += m_coefs[25] * DataPtr[25*2];
		out += m_coefs[26] * DataPtr[26*2];
		out += m_coefs[27] * DataPtr[27*2];
		out += m_coefs[28] * DataPtr[28*2];
		out += m_coefs[29] * DataPtr[29*2];
		out += m_coefs[30] * DataPtr[30*2];
		out += m_coefs[31] * DataPtr[31*2];

		OutBuff[samp][channel] = out;
	} 

	return TRUE;
}
BOOL CAccessSamples::init(PAVISTREAM AudioStream)
/*
This function inits the access object.  It returns TRUE
if successfull.
*/
{
    m_stream = NULL;

    //read audio stream info
    long FmtSize = sizeof(m_FmtData);
    if (0 != AVIStreamReadFormat(AudioStream, 0, &m_FmtData, &FmtSize))
        return FALSE;
	if (m_FmtData.wFormatTag != WAVE_FORMAT_PCM)
        return FALSE;
	else if (m_FmtData.nChannels < 1)
        return FALSE;
	else if (m_FmtData.wBitsPerSample > 16)
        return FALSE;

    m_stream = AudioStream;
    return TRUE;
}

long CAccessSamples::read(SAMPLE *buff, long StartSample, long size)
/*
read the specified number of samples from the specifed location of the
objects stream.  Returns the number of samples read
*/
{
    //exit if stream not set
    if (m_stream == NULL)
        return 0;

    //calc size of raw data & re-alloc buffer if required
    long RawSize = m_FmtData.nBlockAlign * size;
    if (RawSize > m_RawDataSize) {
        m_RawData = (char *)realloc(m_RawData, RawSize);
        if (m_RawData == NULL) {
            m_RawDataSize = 0;
            return 0;
        }
        m_RawDataSize = RawSize;
    }

    //read raw data
    long SamplesRead = 0;
    AVIStreamRead(m_stream, StartSample, size, m_RawData, RawSize, NULL, &SamplesRead);

    //load channel 0 data to sample buffer
    ReadChannel(0, SamplesRead, buff);

    if (m_FmtData.nChannels == 1) {
        //set channel 1  = channel 0 if mono
        long i;
        for (i=0; i< SamplesRead; i++)
            buff[i][1] = buff[i][0];
    }
    else {
        //fill channel 1 with data if stereo
        ReadChannel(1, SamplesRead, buff);
    }

    return SamplesRead;
}

void CAccessSamples::ReadChannel(int channel, long NumSamples, SAMPLE *buff)
/*
this function copies (and converts to float) raw data from the
specified channel to the specifed sample buffer based on the
stream 8/16 bit setting
*/
{
    long i;
    if (m_FmtData.wBitsPerSample > 8) {
        //16 bit case
        char *dataptr = m_RawData + 2*channel;
        for (i = 0; i< NumSamples; i++) {
            buff[i][channel] = (float)*((short *)dataptr);  
            dataptr += m_FmtData.nBlockAlign;
        }
    }
    else {
        //8 bit case
        unsigned char *dataptr = (unsigned char *)m_RawData + channel;
        for (i = 0; i< NumSamples; i++) {
            buff[i][channel] = (float)(((int)*dataptr - 128) << 8);             
            dataptr += m_FmtData.nBlockAlign;
        }
    }
 }

long CAccessSamples::outVert(SAMPLE *buff, long size)
{
    long i;

    //calc size of raw data & re-alloc buffer if required
    long RawSize = m_FmtData.nBlockAlign * size;
    if (RawSize > m_RawDataSize) {
        m_RawData = (char *)realloc(m_RawData, RawSize);
        if (m_RawData == NULL) {
            m_RawDataSize = 0;
            return 0;
        }
        m_RawDataSize = RawSize;
    }
    
    //if stream has more than 2 channels, zero upper channels  
    if (m_FmtData.nChannels > 2)
        memset(m_RawData, 0, RawSize);

    if (m_FmtData.nChannels == 1) {
        //mono case, set data to average of right and left
        if (m_FmtData.wBitsPerSample > 8) {
            //16 bit case
            short *dataptr = (short *)m_RawData;
            for (i=0; i<size; i++) {
				long value = (long)(0.5F * (buff[i][0] + buff[i][1]) );	
                *dataptr++ = min(32767, max(-32767, value));
			}
        }
        else {
            //8 bit case (offset binary)
            unsigned char *dataptr = (unsigned char *)m_RawData;
            for (i=0; i<size; i++) {
				long value = (long)(0.5F * (buff[i][0] + buff[i][1]) );	
                *dataptr++ = 128 + (char)(min(32767, max(-32767, value)) >> 8);
			}
        }
    }
    else {
        //stereo case, copy both channels
        if (m_FmtData.wBitsPerSample > 8) {
            //16 bit case
            char *dataptr = m_RawData;
            for (i=0; i<size; i++) {
				long value = (long)buff[i][0];
                ((short *)dataptr)[0] = (short)min(32767, max(-32767, value));
				value = (long)buff[i][1];
                ((short *)dataptr)[1] = (short)min(32767, max(-32767, value));
                dataptr += m_FmtData.nBlockAlign;
            }
        }
        else {
            //8 bit case (offset binary)
            unsigned char *dataptr = (unsigned char *)m_RawData;
            for (i=0; i<size; i++) {
				long value = (long)buff[i][0];
                dataptr[0] = 128 + (char)(min(32767, max(-32767, value)) >> 8);
				value = (long)buff[i][1];
                dataptr[1] = 128 + (char)(min(32767, max(-32767, value)) >> 8);
                dataptr += m_FmtData.nBlockAlign;
            }
        }
    }
	return RawSize;
}

long CAccessSamples::write(SAMPLE *buff, long StartSample, long size)
/*
Write the specified number of samples to the specifed location of the
objects stream.  Returns the number of samples written
*/
{
    //exit if stream not set
    if (m_stream == NULL)
        return 0;

	long	RawSize = outVert(buff, size);

    //write raw data
    long SamplesWritten = 0;
    AVIStreamWrite(m_stream, StartSample, size, m_RawData, RawSize,
                                    AVIIF_KEYFRAME, &SamplesWritten, NULL);

    return SamplesWritten;
}


BOOL CConvertStream::init(PAVISTREAM AudioStream, double OutputRate)
/*
init convert stream object, return TRUE if all OK.
*/
{
	if (!m_access.init(AudioStream))
		return FALSE;

	m_SameRate = (m_access.GetSampleRate() == OutputRate);
	delete m_RateConvert;
	m_RateConvert = NULL;
	if (!m_SameRate) {
		m_RateConvert = new CConvertRate(m_access.GetSampleRate(), OutputRate);
		if (!m_RateConvert)
			return FALSE;
	}

	m_sample = AVIStreamStart(AudioStream);
	m_OutputRate = OutputRate;
	m_OldSampleValid = FALSE;
	m_ConvDataSize = 0;
	m_ConvData = NULL;

	return TRUE;
}

CConvertStream::~CConvertStream()
/*
distructor
*/
{
	if (m_ConvData)
		free(m_ConvData);
	delete m_RateConvert;
}

long CConvertStream::read(SAMPLE *buff, long size)
/*
read and sample rate convert the next  size (output size) samples from the
the objects avi stream.  Returns the number of sample read.
NOTE:
buff must be at 2*FILTER_LENGTH * input rate / output rate + 1 
sample long.

buff must be at least size+1 samples long!!!!!!!!!!!!!!!!!!!!!!!

*/
{
	long OutSize = 0;

	//exit if nothing to do
	if (!buff || size <= 0)
		return OutSize;

	//if no rate conversion, our job is simple
	if (m_SameRate) {
		long SamplesRead = m_access.read(buff, m_sample, size);
		m_sample += SamplesRead;
		return SamplesRead;
	}

	//output the old sample if valid
	if (m_OldSampleValid) {
		buff[0][0] = m_OldSample[0];
		buff[0][1] = m_OldSample[1];
		size--;
		buff++;
		OutSize++;
	}

	//compute input data requirments
	long ReadSize = m_RateConvert->GetReqInSize(size);

	//re-alloc conversion buffer if required
	if (ReadSize > m_ConvDataSize) {
		m_ConvData = (SAMPLE *)realloc(m_ConvData, ReadSize*sizeof(SAMPLE));
		if (!m_ConvData)
			return OutSize;
		m_ConvDataSize = ReadSize;
	}

	//read stream data
	long SamplesRead = m_access.read(m_ConvData, m_sample, ReadSize);
	m_sample += SamplesRead;
	if (SamplesRead == 0)
		return OutSize;

	//zero-pad if needed
	while (SamplesRead < ReadSize) {
		m_ConvData[SamplesRead][0] = 0.0F;	
		m_ConvData[SamplesRead][1] = 0.0F;
		SamplesRead++;
	}
	
	//convert the sample rate
	long ConvSize = m_RateConvert->Convert(m_ConvData, buff, 0, ReadSize);
	OutSize += ConvSize;
	if (m_access.IsMono()) {
		//copy channel 1 if input is mono
		for (int i = 0; i<ConvSize; i++)
			buff[i][1] = buff[i][0];
	}
	else
		m_RateConvert->Convert(m_ConvData, buff, 1, ReadSize);

	//if outsize is too large, save extra point for next time
	m_OldSampleValid = FALSE;
	if (ConvSize > size) {
		//ASSERT(ConvSize == size + 1);
		m_OldSampleValid = TRUE;
		m_OldSample[0] = buff[0][0];
		m_OldSample[1] = buff[0][1];
	}

	return OutSize;
}

long ConvertAudioSamples(SAMPLE *buff, void *RawOutData, 
						 WAVEFORMATEX * pFmtData, long NumSamples)
/*
This function converts the specified audio samples into the specified uncompressed
format.

The following fields form pFmtData are used:
	nBlockAlign
	nChannels
	wBitsPerSample

*/
{
    long i;

    //calc size of raw data & re-alloc buffer if required
    long RawSize = pFmtData->nBlockAlign * NumSamples;
   
    //if stream has more than 2 channels, zero upper channels  
    if (pFmtData->nChannels > 2)
        memset(RawOutData, 0, RawSize);

    if (pFmtData->nChannels == 1) {
        //mono case, set data to average of right and left
        if (pFmtData->wBitsPerSample > 8) {
            //16 bit case
            short *dataptr = (short *)RawOutData;
            for (i=0; i<NumSamples; i++) {
				long value = (long)(0.5F * (buff[i][0] + buff[i][1]) );	
                *dataptr++ = min(32767, max(-32767, value));
			}
        }
        else {
            //8 bit case (offset binary)
            unsigned char *dataptr = (unsigned char *)RawOutData;
            for (i=0; i<NumSamples; i++) {
				long value = (long)(0.5F * (buff[i][0] + buff[i][1]) );	
                *dataptr++ = 128 + (char)(min(32767, max(-32767, value)) >> 8);
			}
        }
    }
    else {
        //stereo case, copy both channels
        if (pFmtData->wBitsPerSample > 8) {
            //16 bit case
            char *dataptr = (char *)RawOutData;
            for (i=0; i<NumSamples; i++) {
				long value = (long)buff[i][0];
                ((short *)dataptr)[0] = (short)min(32767, max(-32767, value));
				value = (long)buff[i][1];
                ((short *)dataptr)[1] = (short)min(32767, max(-32767, value));
                dataptr += pFmtData->nBlockAlign;
            }
        }
        else {
            //8 bit case (offset binary)
            unsigned char *dataptr = (unsigned char *)RawOutData;
            for (i=0; i<NumSamples; i++) {
				long value = (long)buff[i][0];
                dataptr[0] = 128 + (char)(min(32767, max(-32767, value)) >> 8);
				value = (long)buff[i][1];
                dataptr[1] = 128 + (char)(min(32767, max(-32767, value)) >> 8);
                dataptr += pFmtData->nBlockAlign;
            }
        }
    }
	return RawSize;
}

