#include "stream.h"
#include "..\errors.h"
#include <string.h>
#include <math.h>

//define parameters
#define MIN_PADDING_FRACTION 0.01 //minimum padding as fraction of stream rate
                                  //to account for error in the reported video rate
#define VIDEO_RATE_AVE_SIZE 1.5  //how much data to average to get video rate
                                 //in units of VBV buffer size
#define NUM_AUDIO_BUFFS 20
#define AUDIO_BUFF_SIZE (1152*2)
#define NUM_VIDEO_BUFFS 30
#define VIDEO_BUFF_SIZE 4096

#define USE_RATE_TRACKING            //use code to track video rate so actual delay matches VBV delay
#define RATE_TRACKING_AVE_TIME 2.0   //time the average delay error to get rate error (in seconds)

#define AUDIO_FRAMES_AHEAD 1     //if there will be less then this number of
                                 //audio frames in the decoders buffer after the next 
                                 //video packet is send,  the next packet should be 
                                 // audio data
#define MAX_AUDIO_SAMPLES_AHEAD (8*1024)    //max number of decoded audio samples
                                 //we will be ahead.  The audio packet size will be adjusted
                                 //so we do not exceed this amount (some mpeg players
                                 //don't work well if we exceed this amount, even though
                                 //the std buffer size is in spec)
#define MIN_PACKET_SIZE_FRACTION 0.90  //min packet size as fraction of ave packet size
                                 //used to calc max padding for mAudioTimeAhead calculation
#define MIN_DATA_RATE_FRACTION 0.25  //used for non-fixed rates only:
                                 //min data rate as fraction of 1st rate

#define CSPS_AUDIO_BUFFER_SIZE 4096         //max STD audio buffer for CSPS
#define CSPS_VIDEO_BUFFER_SIZE (46 * 1024)  //max STD video buffer for CSPS

//#define SMALL_FIRST_PACK

//system header data structures and size definitions

struct SSystemHeader {
    char SystemHeaderStartCode[4];
    char HeaderLength[2];
    char MaxMuxRateData[3];
    char flags1;                //audio bound (6 bits), fixed flag, CSPS flag
    char flags2;                //audio lock, video lock, marker, video bound (5 bits)
    char reserved;
    char VideoStreamID;
    char VideoBufferSize[2];
    char AudioStreamID;
    char AudioBufferSize[2];
};
#define SYSTEM_HEADER_SIZE sizeof(SSystemHeader)



#define MIN_PADDING_BYTES (PACKET_HEADER_SIZE+1)  //add one for 0x0f end of header code


void CStreamMix::ProcessData(void)
/*
This is the main processing loop of the MPEG-1 stream mixing class.  It builds and outputs
data packets and packs until it runs out of input data.
*/
{
    CAutoLock AutoLock(&mCritSec);       //don't forget to make thread safe

    if (!mStreamStarted) {
        mStreamStarted = StartUpStream();
        if (!mStreamStarted)
            return;      //exit if inputs are not yet ready
    }

    while (1) {
        double CurOutputRate; 

        //fetch data for next packet
        if (NeedAudio()) {
            if (mAudioPin.ValidData() < mMinAudioSize)
                return;           //return if more audio data needed
            CurOutputRate = GetOutputRate();    //fetch current data rate before data is removed
            BuildAudioPacket();
        }
        else {
            if (mVideoPin.ValidData() < mMinVideoSize)
                return;           //return if more video data needed
            CurOutputRate = GetOutputRate();    //fetch current data rate before data is removed
            BuildVideoPacket();
        }

        ProcessPacket(CurOutputRate);

    } //end while

}

void CStreamMix::FlushData(void)
/*
This function is called when all input data is in the audio and video pin buffers.
This routine roemoves and precesses all remaining data to build the last packets and
packs of the mpeg-1 stream.
*/
{
    //flush video pin input and process "full packets" of data
    mVideoPin.flush();
    ProcessData();

    if (!mStreamStarted)
        return;     //can't flush if we haven't started

    //process remaining data
    int AudioSize = mAudioPin.ValidData();
    int VideoSize = mVideoPin.ValidData();
    while (AudioSize || VideoSize) {

        BOOL GetAudio = AudioSize && (NeedAudio() || !VideoSize); 

        //fetch data for next packet
        if (GetAudio) {
            mCurPacketSize = MIN(mCurPacketSize, AudioSize+AUDIO_PACKET_HEADER_SIZE);
            BuildAudioPacket();
        }
        else {
            mCurPacketSize = MIN(mCurPacketSize, VideoSize+VIDEO_PACKET_HEADER_SIZE);
            BuildVideoPacket(TRUE);
        }

        ProcessPacket(GetOutputRate());

        AudioSize = mAudioPin.ValidData();
        VideoSize = mVideoPin.ValidData();
    } //end while

    //insert stream end code
    unsigned char EndCode[4];
    EndCode[0] = 0;
    EndCode[1] = 0;
    EndCode[2] = 1;
    EndCode[3] = 0xb9;

    unsigned long OutSize;
    WriteFile(mOutputFile, EndCode, sizeof(EndCode), &OutSize, NULL);
    if ((int)OutSize != sizeof(EndCode))
        ProcessError(ERROR_FILE_WRITE);


}

void CStreamMix::ProcessPacket(double CurOutputRate)
/*
This routine writes the current packet data (mPacketData) to the objects
output file.  It then sets up the object's member data for the next
packet.  This function uses the specifed data rate to update
the current stream time.
*/
{
    //output packet data
    unsigned long OutSize;
    WriteFile(mOutputFile, mPacketData, mCurPacketSize, &OutSize, NULL);
    if ((int)OutSize != mCurPacketSize)
        ProcessError(ERROR_FILE_WRITE);
    mUsedPackBytes += mCurPacketSize;

    //update packet time and packet counts
    mStreamTime += mCurPacketSize / CurOutputRate;
    mCurPacketNum++;

    if (mSkipAudioPacketCount > 0 && mCurPacketNum == mPacketsPerPack - 1) {
        //mSkipAudioPacketCount is only used when we are not in just-in-time
        //audio mode.  It is used to aviod stuffing audio in the first few
        //packs so that we can minimize the buffering required befor the start time
        mSkipAudioPacketCount--;
        mCurPacketNum++;
    }

    //if last packet in pack, finish this pack, and sutup next pack
    if (mCurPacketNum == mPacketsPerPack) {
        //add padding packet
        int PaddingSize = OutputPadding(mPackSize - mUsedPackBytes);
        mStreamTime += PaddingSize / GetOutputRate();
        //set counters for next pack
        mUsedPackBytes = mCurPacketNum = 0;
        //output header for next pack        
        OutputPackHeader();
        mUsedPackBytes += PACK_HEADER_SIZE;
        mStreamTime += PACK_HEADER_SIZE / GetOutputRate();
    }

    //set size for next packet
    mCurPacketSize = CalcPacketSize();


}

int CStreamMix::CalcPacketSize(void)
/*
this function computes the size of the current packet as specified by 
member data mCurPacketNum, mPacketsPerPack, mUsedPackBytes, and 
mPackSize  
*/
{
    int PacketSize = mBasePacketSize;

    //special processing for last packet in pack
    if (mCurPacketNum == mPacketsPerPack - 1) {
        if (!mJustInTimeAudio) {
            //if this is a "small" audio packet, set it's size
            PacketSize = (int)(mAveAudioPacketSize + mAudioBytesUnder);
        }

        //insure that the last packet isn't too big
        PacketSize = min(PacketSize, mPackSize - mUsedPackBytes);

        //if there is not enough bytes for padding, fill data to end of pack
        if (mPackSize - (mBasePacketSize + mUsedPackBytes) < MIN_PADDING_BYTES) {
            PacketSize = mPackSize - mUsedPackBytes;
        }

        if (!mJustInTimeAudio) {
            //if "small" audio packet, compute running rate error
            mAudioBytesUnder += mAveAudioPacketSize - PacketSize;
        }
    }

    return PacketSize;
}


void CStreamMix::BuildVideoPacket(BOOL DisableTimeStamp/*=FALSE*/)
/*
This function fills the objects packet data buffer with the next packet of video data.
The function returns the packet size.
*/
{

    //copy video packet header
    char *PacketPtr = mPacketData;
    BuildPacketHeader(PacketPtr, VIDEO_STREAM, mCurPacketSize);
    PacketPtr += PACKET_HEADER_SIZE;
    int DataSize = mCurPacketSize - PACKET_HEADER_SIZE - VIDEO_STAMP_SIZE;

    
    // check for video start code
    if (mVideoPin.RemainingFrameSize() < DataSize && !DisableTimeStamp) {
        //adjust stream data rate to avoid overflow/underflow
        AdjustRates();
        //video start code in this packet, add time stamp  
        AddVideoTimeStamp(PacketPtr);
    }
    else {
        //no video start code in this packet, add stuffing and standard buff size
        //printf("\tVideo: size %d\n", mCurPacketSize);
        int StuffSize = VIDEO_STAMP_SIZE - STD_BUFF_DATA_SIZE - END_CODE_SIZE;
        AddStuffing(PacketPtr, StuffSize, FALSE);
        LoadStdBuffSize(&PacketPtr[StuffSize], VIDEO_STREAM);
        PacketPtr[StuffSize + STD_BUFF_DATA_SIZE] = (char)0x0f;  //mark end of header

    }
    PacketPtr += VIDEO_STAMP_SIZE;

    //copy video data to packet buffer
    mVideoPin.MoveData(PacketPtr, DataSize);

}


int CStreamMix::NeedAudio(void)
/*
This function returns true (1) if an audio packet is required.  It
returns false (0) if a video packet is required.
*/
{
	// if audio input is done, and all video buffers all full, force video processing
	if (mAudioPin.IsInputFinished() && !mVideoPin.BufferReady())
		return FALSE;
 
 	// if video input is done, and all audio buffers all full, force audio processing
	if (mVideoPin.IsInputFinished() && !mAudioPin.BufferReady())
		return TRUE;

    if (mJustInTimeAudio) {
        //just-in-time audio mode, check times
        double TranTime = AudioFrameToTime(mAudioPin.GetNextFrameNumber()) - mAudioTimeAhead;
        return (mStreamTime >= TranTime);
    }
    else {
        //single "small" audio packet per pack mode, is it the last packet?
        return (mCurPacketNum == mPacketsPerPack - 1);
    }
}

void CStreamMix::BuildAudioPacket(void)
/*
This function fills the objects packet data buffer with the next packet of audio data.
The function returns the packet size.
*/
{
    //copy audio packet header
    char *PacketPtr = mPacketData;
    BuildPacketHeader(PacketPtr, AUDIO_STREAM, mCurPacketSize);
    PacketPtr += PACKET_HEADER_SIZE;
    int DataSize = mCurPacketSize - AUDIO_PACKET_HEADER_SIZE;

#if 0
    //make sure we don't send too much (decoded) data
    double CurTimeAread = AudioFrameToTime(mAudioPin.GetNextFrameNumber()) - mStreamTime;
    double CurFramesAhead = CurTimeAread * mAudioFrameRate;
    int MaxSamples = MAX_AUDIO_SAMPLES_AHEAD - (int)(CurFramesAhead * AUDIO_BUFF_SIZE);
    int MaxSize = (int) (MaxSamples * mAudioPin.GetDataRate() / 44100);
    if (DataSize > (MaxSize + (int)MIN_PADDING_BYTES) ) {
        if ((DataSize - MaxSize) >= 100) {
            //if there is enough unused room, send an extra packet
            mCurPacketNum--;
        }
        DataSize = MaxSize;
        mCurPacketSize = DataSize + PACKET_HEADER_SIZE;
    }
#endif

    //does a new audio frame start in this buffer?
    if (mAudioPin.RemainingFrameSize() < DataSize) {
        //if so, add std buffer size & time stamp
        LoadStdBuffSize(PacketPtr, AUDIO_STREAM);
        int FrameNum = mAudioPin.GetNextFrameNumber();
        AddTimeStamp(&PacketPtr[STD_BUFF_DATA_SIZE], SINGLE_STAMP, AudioFrameToTime(FrameNum));
        /* debug print
        double delta = AudioFrameToTime(FrameNum)
                    -    mStreamTime
                    -    mAudioPin.RemainingFrameSize()/mStreamRate;
        printf("\tAudio: size %d, time %f frame %d delta %f\n",
                    mCurPacketSize, AudioFrameToTime(FrameNum), FrameNum, delta);  */

    }
    else {
        //add stuffing if we don't need time stamp
        AddStuffing(PacketPtr, TIME_STAMP_SIZE + STD_BUFF_DATA_SIZE);
//        printf("\tAudio: size %d\n", mCurPacketSize);
    }

    // update pointer
    PacketPtr += TIME_STAMP_SIZE + STD_BUFF_DATA_SIZE;

    //fill packet with audio data
    mAudioPin.MoveData(PacketPtr, DataSize);
}

void CStreamMix::AdjustRates(void)
/*
This function is called when a new frame is encountered.  It computes the video
stream rate that is required to achieve the VBV buffer fullness that is specifed
for the next frame.  
*/
{
    if (!mFixedRate)
        return;  //nothing to do if not fixed rate

    //if 1st frame, init stream start times & exit
    int FrameNum = mVideoPin.GetNextFrameNumber();
    if (FrameNum == 0) {
        double ExtraVideoTime = 0.0;
        if (mJustInTimeAudio) {
            //acount for fact that audio isn't sent at first
            double delay = mVideoPin.GetNextVBVDelay() + mVBVPaddingTime + mAudioDelay - mAudioTimeAhead;
            double ExtraVideoBytes = delay * mAudioPin.GetDataRate();
            ExtraVideoTime = ExtraVideoBytes / (GetOutputRate() + mAudioPin.GetDataRate());
        }

        mVideoStartTime = mVideoPin.GetNextVBVDelay() + mVBVPaddingTime - ExtraVideoTime;
        mAudioStartTime = mVideoStartTime + mAudioDelay;
        mCurVideoFrameSize = mVideoPin.GetNextFrameSize();  //init current frame size

        //init number of initial packs to skip audio
        mSkipAudioPacketCount = 0;
        if (!mJustInTimeAudio) {
            //set to fill 1/2 of the available buffer
            int StartPacket = (int)(GetOutputRate() * mAudioStartTime / mPackSize);
            int PacketsAhead = (int)(0.5 + (CSPS_AUDIO_BUFFER_SIZE/2) / mAveAudioPacketSize);
            mSkipAudioPacketCount = max(0, StartPacket - PacketsAhead);
        }
    /*debug prints
        printf("Frame %4d,Delay (cur next) %f %f, Size (packet frame) %d %d\n",
                FrameNum, 0.0, NextDelay, mBasePacketSize, mCurVideoFrameSize); */
        return;
    }

#ifdef USE_RATE_TRACKING
    //get current delay
    double CurDelay = VideoFrameToTime(FrameNum)
                    -    mStreamTime
                    -    mVideoPin.RemainingFrameSize()/mStreamRate;
    double TargetDelay = mVideoPin.GetNextVBVDelay() + mVBVPaddingTime;
    double RateError = (TargetDelay - CurDelay) * mVideoPin.GetDataRate() / RATE_TRACKING_AVE_TIME;
    mBasePacketSize = (int)(mAvePacketSize + RateError / mPacketRate);
    mBasePacketSize = min(mBasePacketSize, mMaxPacketSize);
    
#else
    mBasePacketSize = mAvePacketSize;
#endif

    /*debug prints  
    int AveFrame = (int)(mVideoPin.GetDataRate()*(1/29.97));
    int CurBuff = (int)((CurDelay - mVBVPaddingTime) * mVideoPin.GetDataRate());
    int vbv = (int)(mVideoPin.GetNextVBVDelay() * mVideoPin.GetDataRate());
    int delta =  AveFrame - mVideoPin.GetNextFrameSize();
    printf("Frame %4d, size %5d, cur buff %5d vbv %5d delta %5d next %5d\n",
                mVideoPin.GetNextFrameNumber(), mVideoPin.GetNextFrameSize(),
                CurBuff, vbv, delta, vbv+delta);

    printf("Frame %4d,Delay(cur next) %f %f, Size(packet frame) %d %d\n",
                FrameNum, CurDelay, TargetDelay, mBasePacketSize, mCurVideoFrameSize);
    */

    mCurVideoFrameSize = mVideoPin.GetNextFrameSize();  //update current frame size

}


void CStreamMix::AddTimeStamp(char * OutPtr, ETimeStampType type, double time)
/*
This functon writes a time stamp of the specified type at the
specified time (given in seconds) to the specified memory buffer.
*/
{
    SStamp * StampPtr = (SStamp *)OutPtr;

    //convert to time in 90K clock ticks
    double two33 = 65536;
    two33 *= 65536*2;  //get 2^33 without compiler constant calculation error
    double ticks = fmod((time * 90000), two33);
    
    //get high (32-30) and low (29-0) bits  
    int HighBits = (int)(ticks / (32768 * 32768));  //get bits 32-30 (/ by 2^30)
    int LowBits = (int)(ticks - HighBits * 32768 * 32768 + .5);  //round to get bits 29-0

    //write output
    StampPtr->TypeByte = (type << 4 ) + (HighBits << 1) + 1; //type, highbits, and marker bit
    unsigned short MidWord = (LowBits >> (15-1)) | 1;  //store bits 29-15 and marker bit
    StampPtr->MidWord[0] = (unsigned char)(MidWord >> 8);
    StampPtr->MidWord[1] = (unsigned char)MidWord;
    unsigned short LowWord = (LowBits << 1) | 1;          //store bits 14-0 and marker bit
    StampPtr->LowWord[0] = (unsigned char)(LowWord >> 8);
    StampPtr->LowWord[1] = (unsigned char)LowWord;
} 


int CStreamMix::AddVideoTimeStamp(char * OutPtr, BOOL Pad /*=TRUE */)
/*
This funciton adds a video time stamp (presentation stamp folowed by 
decoding stamp) to the specified output buffer.
*/
{
    char * DataPtr = OutPtr;
    int FrameNum = mVideoPin.GetNextFrameNumber();
    int PresNum = mVideoPin.GetNextPresentationNumber() + 1; //note:  we add 1 so
                                                             //presentation numbers start at
                                                             //1.  this insures that decode
                                                             //numbers are always <= presentation
                                                             //numbers

/*  print out for debug 
    double delta = VideoFrameToTime(FrameNum)
                    -    mStreamTime
                    -    mVideoPin.RemainingFrameSize()/mStreamRate;
    printf("\tVideo: size %d, time %f frame %d pres %d delta %f\n", mCurPacketSize,
                      VideoFrameToTime(FrameNum), FrameNum, PresNum, delta);*/

    if (FrameNum == PresNum) {
        //if presentation time == decode time, only encode presentation time
        if (Pad)
		{
			AddStuffing(DataPtr, TIME_STAMP_SIZE, FALSE);
			DataPtr += TIME_STAMP_SIZE;
		}
        AddTimeStamp(DataPtr, SINGLE_STAMP, VideoFrameToTime(PresNum));
		DataPtr += TIME_STAMP_SIZE;
    }   
    else {
        //if presentation time != decode time, presentation decode time & encode time
        AddTimeStamp(DataPtr, FIRST_STAMP, VideoFrameToTime(PresNum));
        DataPtr += TIME_STAMP_SIZE;
        AddTimeStamp(DataPtr, SECOND_STAMP, VideoFrameToTime(FrameNum));
		DataPtr += TIME_STAMP_SIZE;
    }
	return DataPtr-OutPtr;
}

void CStreamMix::AddStuffing(char * OutPtr, int size, BOOL EndHeader/*=TRUE*/)
/*
this funciton adds the specified number of stuffing bytes to
the specified output buffer.  If end header flag is set, the function
outputs the 0x0f marker to signify the end of the packet header as the
last suffing byte
*/
{

    int i;
    int sz = size;
    if (EndHeader)
        sz--;

    for (i=0; i<sz; i++)
        *OutPtr++ = (char)0xff;

    if (EndHeader && size)
        *OutPtr = (char)0x0f;
}

BOOL CStreamMix::StartUpStream(void) 
/*
This function is called to start up the stream.  It initializes all required
data.  The function returns TRUE if successfull, FALSE if more input data is 
required form the audio or video pins
*/
{
    //compute stream rate parameters
    if (!InitRateData(MIN_PADDING_FRACTION))
        return FALSE;

    //allocate packet data storage
    mPacketData = new char[mMaxPacketSize];
    
    //init stream start times
    if (mFixedRate) {
        //init start times so we get video packet 1st,
        //will fill in start time when we get the 1st video frames VBV delay
        //calc stream start times
        mVideoStartTime = 0.0;
        mAudioStartTime = 100000000.0;
    }
    else {
        if (mVideoPin.ValidData() < mMinVideoSize)
                return FALSE;         //return if more video data needed to average data rate
        mVideoStartTime = GetVideoBufferSize() / GetOutputRate();
        mAudioStartTime = mVideoStartTime + mAudioDelay;
    }

    //init time delta to fill audio (used by NeedAudiio() )
    double MinDataRate = mFixedRate ? GetOutputRate() : GetOutputRate() * MIN_DATA_RATE_FRACTION;
    int MinPacketSize = (int)(MIN_PACKET_SIZE_FRACTION * mAvePacketSize);
    int MaxNonDataBytes = mPackSize - mPacketsPerPack * MinPacketSize;
    mAudioTimeAhead = AUDIO_FRAMES_AHEAD / mAudioFrameRate 
                    + (MaxNonDataBytes + mMaxPacketSize) / MinDataRate;

        
#ifdef SMALL_FIRST_PACK
    //output pack header of 1st pack with system header
    mStreamTime = 0.0;
    int FirstSize = OutputFirstPack();
    mStreamTime += FirstSize / GetOutputRate();

    //set up for 1st packet
    mUsedPackBytes = SYSTEM_HEADER_SIZE;
    mCurPacketNum = 0;
    mStreamTime +=  SYSTEM_HEADER_SIZE / GetOutputRate();
    mCurPacketSize = CalcPacketSize();
#else
    //output pack header of 1st pack with system header
    mStreamTime = 0.0;
    OutputPackHeader();
    OutputSystemHeader();

    //set up for 1st packet
    mUsedPackBytes = PACK_HEADER_SIZE + SYSTEM_HEADER_SIZE;
    mCurPacketNum = 0;
    mStreamTime +=  (PACK_HEADER_SIZE + SYSTEM_HEADER_SIZE) / GetOutputRate();
    mCurPacketSize = CalcPacketSize();
    //small 1st packet to account for system header
    mCurPacketSize -= SYSTEM_HEADER_SIZE; 
#endif
    return TRUE;
}



BOOL CStreamMix::InitRateData(double MinPaddingFraction) 
/*
This function is called to set up all stream rate / buffer size data values.
The function returns TRUE if successfull, FALSE if more input data is 
required form the audio or video pins.
The MinPaddingFraction parameter sets the min padding as a fraction of stream rate
*/
{
    //exit if audio and video stream headers info are not yet available
    if (!mAudioPin.IsHeaderReady() || !mVideoPin.IsHeaderReady())
        return FALSE;
        
    //fetch audio and video rates
    mAudioFrameRate = mAudioPin.GetFrameRate();
    if (mVideoFrameRate == 0.0) {
        //if rate not supplied in Initialize(), use stream rate
        mVideoFrameRate = mVideoPin.GetFrameRate();
    }
    double AudioRate = mAudioPin.GetDataRate();
    double VideoRate = mVideoPin.GetDataRate();
    mFixedRate = VideoRate != 0.0;  //VideoRate of 0.0 ==> non-fixed rate operation

    //init packet size and stream rate variables
    if (mFixedRate) {
        //calculate ave packet size & stream rates based on fixed rate
        double DataRate = AudioRate + VideoRate;
        double PackDataBytes;
        double PacketOverhead;
        if (mJustInTimeAudio) {
            PacketOverhead = AudioRate/DataRate * AUDIO_PACKET_HEADER_SIZE
                           + VideoRate/DataRate * VIDEO_PACKET_HEADER_SIZE;
            PackDataBytes = mPackSize - PACK_HEADER_SIZE - PacketOverhead * mPacketsPerPack;
        }
        else {
            PackDataBytes = mPackSize - PACK_HEADER_SIZE
                                      - AUDIO_PACKET_HEADER_SIZE
                                      - VIDEO_PACKET_HEADER_SIZE;
        }

        double ReqRate = DataRate * mPackSize / PackDataBytes;
        mOverHeadRate = ReqRate - DataRate;

        double MinRate = ReqRate * (1.0 + MinPaddingFraction);
        if (mStreamRate == 0.0)
            mStreamRate = MinRate;                 //if requested rate is 0, use min required
        else if (mStreamRate < MinRate) {
            ProcessError(ERROR_RATE_TOO_SLOW);     //error if requested rate too slow
            return FALSE;
        }

        mPacketRate = mPacketsPerPack * mStreamRate / mPackSize;
        mMaxStreamRate = mStreamRate;

        if (mJustInTimeAudio) {
            mBasePacketSize = (int)(0.5 + PacketOverhead + DataRate/mPacketRate);
        }
        else {
            double PackRate = mStreamRate / mPackSize;
            mAveAudioPacketSize = AudioRate / PackRate + AUDIO_PACKET_HEADER_SIZE;
            mAudioBytesUnder = 0.0;
            mBasePacketSize = (int) (0.5 + VideoRate / PackRate + VIDEO_PACKET_HEADER_SIZE);
        }

        //convert VBV padding to time
        mVBVPaddingTime = mVBVPadding / VideoRate;

    }
    else {
        //not fixed rate, set packet size = available bytes
        mBasePacketSize = (mPackSize - PACK_HEADER_SIZE) / mPacketsPerPack;
		mAveAudioPacketSize = mBasePacketSize;
        mMaxStreamRate = 0.0;
    }

    //compute packet size bounds:
    // max = min padding bytes + PackData/mPacketsPerPack (round up)
    int PackData = mPackSize - PACK_HEADER_SIZE;
    if (mJustInTimeAudio)
        mMaxPacketSize = MIN_PADDING_BYTES + (PackData + mPacketsPerPack - 1) / mPacketsPerPack; 
    else {
		int AudioSize = (int)ceil(mAveAudioPacketSize);
        mMaxPacketSize = max(AudioSize, PackData - AudioSize);
	}
    mAvePacketSize = mBasePacketSize;

    //init stream input requirments
    mMinAudioSize = mMaxPacketSize;
    if (mFixedRate)
        mMinVideoSize = mMaxPacketSize;
    else {
        mMinVideoSize = (int) (VIDEO_RATE_AVE_SIZE * mVideoPin.GetVBVSize());
        mMinVideoSize = max(mMinVideoSize, mMaxPacketSize);
    }

	return TRUE;
}


int CStreamMix::Initialize( HANDLE OutputFile,    //win32 file handle
                            int PackSize,         //in bytes
                            int PacketsPerPack,   //if 0, we will use 2 packets per pack
                                                  //the first will be an video packet, the 
                                                  //second an audio packet.  If this is non-zero
                                                  //we use the specified number of same
                                                  //size packets in just-in-time audio mode.
                                                  //in this case audio is sent just before the
                                                  //audio buffers underflow.
                            double StreamRate,    //rate of output stream in bytes/second
                                                  //(if set to zero, min required rate
                                                  //will be used.  Ignored for non-fixed rate video)
                            double AudioDelay,    //delay audio form video presentation in seconds
                                                  //(can be negative)
                            int    VBVPadding,    //extra buffer padding size in bytes
                            double VideoFrameRate) //in frames/second, 0.0 => use rate as specified
                                                    //video stream
/*
This function initializes the CStreamMix object.  It returns 0 if no error, otherwise it ruturs 
the appropriate error code
*/
{

	//if PacketsPerPack != 0, use just-in-time audio 
	BOOL JitAudio = TRUE;
    if (PacketsPerPack == 0) { 
	   //not just-in-time: 1 audio and 1 video packet in each pack
       JitAudio = FALSE;
       PacketsPerPack = 2;
    }

    if (PackSize < PACK_HEADER_SIZE + SYSTEM_HEADER_SIZE)
        return ERROR_PACK_SIZE_TOO_SMALL;
    if ((PackSize-PACK_HEADER_SIZE)/PacketsPerPack <= VIDEO_PACKET_HEADER_SIZE)
        return ERROR_PACKET_SIZE_TOO_SMALL;
    if (!mAudioPin.initialize(NUM_AUDIO_BUFFS, AUDIO_BUFF_SIZE))
        return ERROR_STARTUP_ALLOCATION;
    if (!mVideoPin.initialize(NUM_VIDEO_BUFFS, VIDEO_BUFF_SIZE))
        return ERROR_STARTUP_ALLOCATION;

    mOutputFile = OutputFile;
    mPackSize = PackSize;
    mStreamRate = StreamRate;
    mAudioDelay = AudioDelay;
    mVBVPadding = VBVPadding;
    mVideoFrameRate = VideoFrameRate;
    mStreamStarted = FALSE;
    mJustInTimeAudio = JitAudio;
    mPacketsPerPack = PacketsPerPack;

    return 0;
}

double CStreamMix::GetOutputRate(void)
/*
This function  returns the current output data rate.
If the stream rate is not fixed, the funciton
also updates the stream rate and max stream rate member variables.
*/
{
    //calc new data rate if we are in non-fixed mode and have enough data
    if (!mFixedRate && mVideoPin.ValidData() >= mMinVideoSize) {
        double AudioRate = mAudioPin.GetDataRate();
        double VideoRate = mVideoPin.GetAveFrameSize(mMinVideoSize) * mVideoFrameRate;
        double DataRate = AudioRate + VideoRate;

        double PackDataBytes;
        if (mJustInTimeAudio) {
            double DataRate = AudioRate + VideoRate;
            double PacketOverhead = AudioRate/DataRate * AUDIO_PACKET_HEADER_SIZE
                                  + VideoRate/DataRate * VIDEO_PACKET_HEADER_SIZE;
            PackDataBytes = mPackSize - PACK_HEADER_SIZE - PacketOverhead * mPacketsPerPack;
        }
        else {
            PackDataBytes = mPackSize - PACK_HEADER_SIZE
                                      - AUDIO_PACKET_HEADER_SIZE
                                      - VIDEO_PACKET_HEADER_SIZE;
        }
        mStreamRate = DataRate * mPackSize / PackDataBytes;

        //update max data rate
        mMaxStreamRate = MAX(mMaxStreamRate, mStreamRate);
    }

    return mStreamRate;
}


void CStreamMix::BuildPacketHeader(char *HeaderData, EPacketType PacketType, int NumBytes)
/*
This function builds a pack header as specified in the specified output buffer.  Note
that NumBytes refers to the overall packet size including header
*/
{
    SPacketHeader *header = (SPacketHeader *)HeaderData;

    header->PacketStartCode[0] = 0;
    header->PacketStartCode[1] = 0;
    header->PacketStartCode[2] = 1;

    if (PacketType == PADDING_STREAM) {
        header->StreamID = (char)0xbe;
    }
    else if (PacketType == AUDIO_STREAM) {
        header->StreamID = (char)0xc0;
    }
    else {//VIDEO_STREAM
        header->StreamID = (char)0xe0;
    }

    int length = NumBytes - sizeof(SPacketHeader);
    header->PacketLength[0] = (length >> 8) & 0xff;
    header->PacketLength[1] = length & 0xff;

}

void CStreamMix::LoadStdBuffSize(char *StdSizeData, EPacketType PacketType)
/*
This function writes the 2 byte standard buffer size field into the specified 
location based on the specified pack type.
*/
{
    int SizeBits;
    if (PacketType == PADDING_STREAM) {
        SizeBits = 0x0000ffff;
    }
    else if (PacketType == AUDIO_STREAM) {
        SizeBits = PackBufferSize(GetAudioBufferSize()) | 0x00004000;
    }
    else {//VIDEO_STREAM
        SizeBits = PackBufferSize(GetVideoBufferSize(),TRUE) | 0x00004000;
    }

    StdSizeData[0] = (char)(SizeBits >> 8);
    StdSizeData[1] = (char)(SizeBits);

}

int CStreamMix::OutputPadding(int NumBytes)
/*
This funciton writes a padding packet of the specified size to the
object's stream output file.  If the requested size is smaller then
the smallest possible padding packet, no padding is added.  The funciton
returns the number of padding bytes actualy writen
*/
{
//    printf("\tPadding: size %d\n",NumBytes);

    SPacketHeader header;
    char PadData[64];
    unsigned long OutSize;

    if (NumBytes <MIN_PADDING_BYTES)
        return 0;

    //output header for padding data 
    BuildPacketHeader((char *)&header, PADDING_STREAM, NumBytes);
    WriteFile(mOutputFile, &header, PACKET_HEADER_SIZE, &OutSize, NULL);
    if (OutSize != PACKET_HEADER_SIZE) {
        ProcessError(ERROR_FILE_WRITE);
        return 0;
    }

    int length = NumBytes - PACKET_HEADER_SIZE;

    //init padding data array  (1st byte 0x0f the rest 0xff)
    int i;
    int size = min(length, sizeof(PadData));
    PadData[0] = (char)0x0f;
    for (i=1; i<size; i++)
        PadData[i] = (char)0xff;

    //output padding data
    while (length) {
        size = min(length, sizeof(PadData));
        WriteFile(mOutputFile, PadData, size, &OutSize, NULL);
        if ((int)OutSize != size) {
            ProcessError(ERROR_FILE_WRITE);
            return NumBytes - length - PACKET_HEADER_SIZE;
        }
        length -= size;
        PadData[0] = (char)0xff;  //only want 0x0f on 1st padding byte
    }

    return NumBytes;
}

void CStreamMix::OutputPackHeader(void)
/*
This function writes a pack header to the object's stream output file.
*/
{
    SPackHeader header;

    //load sync code
    header.PackStartCode[0] = 0;
    header.PackStartCode[1] = 0;
    header.PackStartCode[2] = 1;
    header.PackStartCode[3] = (char)0xba;

    //calc time of last byte of stream time field & fill in time stamp
    int TimeByte = sizeof(header.PackStartCode) + sizeof(header.StreamTime) -1;
    double time = mStreamTime + TimeByte / GetOutputRate();
    AddTimeStamp(header.StreamTime, SINGLE_STAMP, time);

    //set the stream data rate for this pack
    int MuxRate = (int)ceil(GetOutputRate()/50.0);   //in units of 50 bytes/second (round up)
    MuxRate &= 0x003fffff;  //22 data bits
    MuxRate = (MuxRate << 1) | 0x00800001;  //set marker bits
    header.MuxRate[0] = (char)(MuxRate >> 16);
    header.MuxRate[1] = (char)(MuxRate >> 8);
    header.MuxRate[2] = (char)MuxRate;

    //write header to file
    unsigned long OutSize;
    WriteFile(mOutputFile, &header, PACK_HEADER_SIZE, &OutSize, NULL);
    if (OutSize != PACK_HEADER_SIZE)
        ProcessError(ERROR_FILE_WRITE);

//    printf("-------- Pack time %f\n", time);
}

int CStreamMix::OutputFirstPack(void)
/*
This function outputs the 1st pack.  This includes no stream data but does 
contaion the system header and the 1st packet of each stream that is used to
set the STD buffer size.  The function returns the size of the first pack.
*/
{
    struct SFirstPacket {
        SPacketHeader header;
        char StdBuffSize[2];
        char EndCode;
    } packet;

    OutputPackHeader();
    OutputSystemHeader();

    //built 1st video packet
    BuildPacketHeader((char *)&packet.header, VIDEO_STREAM, sizeof(packet));
    //add video stream STD buffer info
    LoadStdBuffSize(packet.StdBuffSize, VIDEO_STREAM);
    packet.EndCode = 0x0f;

    //write packet to file
    unsigned long OutSize;
    WriteFile(mOutputFile, &packet, sizeof(packet), &OutSize, NULL);
    if (OutSize != sizeof(packet))
        ProcessError(ERROR_FILE_WRITE);


    //built 1st audio packet
    BuildPacketHeader((char *)&packet.header, AUDIO_STREAM, sizeof(packet));
    LoadStdBuffSize(packet.StdBuffSize, VIDEO_STREAM);
    packet.EndCode = 0x0f;

    //write packet to file
    WriteFile(mOutputFile, &packet, sizeof(packet), &OutSize, NULL);
    if (OutSize != sizeof(packet))
        ProcessError(ERROR_FILE_WRITE);

    return PACK_HEADER_SIZE+SYSTEM_HEADER_SIZE + 2* sizeof(packet);

}

void CStreamMix::OutputSystemHeader(void)
/*
This function writes the system header to the object's stream output file.
*/
{
    SSystemHeader header;

    //load sync code
    header.SystemHeaderStartCode[0] = 0;
    header.SystemHeaderStartCode[1] = 0;
    header.SystemHeaderStartCode[2] = 1;
    header.SystemHeaderStartCode[3] = (char)0xbb;

    //load header length
    int length = SYSTEM_HEADER_SIZE - sizeof(header.SystemHeaderStartCode)
                                    - sizeof(header.HeaderLength);
    header.HeaderLength[0] = (char)(length >> 8);
    header.HeaderLength[1] = (char)length;

    //load max stream rate
    //note: for varibale rate streams, we need to fill in this field 
    //      at the end of file once we know the max rate.
    int MuxRate = (int)ceil(mMaxStreamRate/50.0);   //in units of 50 bytes/second (round up)
    MuxRate &= 0x003fffff;  //22 data bits
    MuxRate = (MuxRate << 1) | 0x00800001;  //set marker bits
    header.MaxMuxRateData[0] = (char)(MuxRate >> 16);
    header.MaxMuxRateData[1] = (char)(MuxRate >> 8);
    header.MaxMuxRateData[2] = (char)MuxRate;

    //load num audio streams=1 (6 bits), fixed flag, CSPS flag
    header.flags1 = (1<<2) + (mFixedRate << 1) + IsCSPS();

    //load audio lock=1, video lock=1, marker=1, num video streams=1(5 bits)
    header.flags2 = (char)0xe1;

    header.reserved = (char)0xff;

    //video stream max STD buffer info
    header.VideoStreamID = (char)0xe0; 
    int SizeBits = PackBufferSize(GetVideoBufferSize(),TRUE) | 0x0000c000;
    header.VideoBufferSize[0] = (char)(SizeBits >> 8);
    header.VideoBufferSize[1] = (char)SizeBits;

    //audio stream max STD buffer info
    header.AudioStreamID = (char)0xc0; 
    SizeBits = PackBufferSize(GetAudioBufferSize()) | 0x0000c000;
    header.AudioBufferSize[0] = (char)(SizeBits >> 8);
    header.AudioBufferSize[1] = (char)SizeBits;

    //write header to file
    unsigned long OutSize;
    WriteFile(mOutputFile, &header, SYSTEM_HEADER_SIZE, &OutSize, NULL);
    if (OutSize != SYSTEM_HEADER_SIZE)
        ProcessError(ERROR_FILE_WRITE);

}

int CStreamMix::GetVideoBufferSize(void)
/*
This function returns the size of the decode's video buffer required to process
the stream produced by this object
*/
{
    //ASSERT(mStreamStarted);

    //note: report 2 * VBVPadding, gives VBVPadding bytes for both
    //      underflow and overflow protection
    return mVideoPin.GetVBVSize() + 2*mVBVPadding;
}

int CStreamMix::GetAudioBufferSize(void)
/*
This function returns the size of the decode's audio buffer required to process
the stream produced by this object
*/
{
    if (!mJustInTimeAudio)
        return CSPS_AUDIO_BUFFER_SIZE;

    //compute max bytes needed by audio data = #audio bytes to stay ahead + packet size
    int AudioBytesAhead = (int)(mAudioTimeAhead * mAudioPin.GetDataRate());
    int MaxAudioBytes = AudioBytesAhead + mMaxPacketSize + MIN_PADDING_BYTES;

    return MaxAudioBytes;
}

BOOL CStreamMix::IsCSPS(void)
/*
This funciton returns TRUE if the objects parameters meet the ISO/IEC 11172-1
constraind system parameter requirments.  It returns FALSE otherwise.
*/
{
    //packet rate must be <= 300 /second OR packet size >= 50000/300/8
    if (mPacketRate > 300.0 && (double)mBasePacketSize < 5000000.0/300/8 )
        return FALSE;

    //audio buffer size must be <= 4096
    if (GetAudioBufferSize() > CSPS_AUDIO_BUFFER_SIZE)
        return FALSE;

    //video buffer size must be <= 46 * 1024
    if (GetVideoBufferSize() > CSPS_VIDEO_BUFFER_SIZE)
        return FALSE;

    return TRUE;
}

int CStreamMix::PackBufferSize(int BuffSize,BOOL VideoScale/*=FALSE*/)
/*
This function packs the specified buffer size into a 16 bit number
according to ISO/IEC 11172-1 sec 2.4.4.2:

d15-d14 = 11 binary
d13 = 1, units are 1024 bytes, d13 = 0, units are 128 bytes
d12-d0 buffer size: in units of 128 or 1024 based on d13

*/
{
    int scale = 0;
    int size = (BuffSize + 127)/128;
    if (size > 0x01ff || VideoScale) {
        scale = 1;
        size = (BuffSize + 1023)/1024;
    }

    return (scale << 13) + size;
}





