/*
 * vcinstal.c
 *
 * 32-bit Video Capture driver
 * User-mode support library - kernel driver install/remove
 *
 *
 * Mark Vrabel, October 95:
 */

#include <vcdriver.h>
#include <vtdioc.h>
#include "vcupriv.h"
#include "registry.h"

/*
 * Driver Install/Remove
 *
 * we use the standard functions in registry.c to start and
 * stop the kernel driver, and to read/write the registry for
 * driver configuration information. This module is essentially a
 * wrapper to permit common win-95/nt driver code.
 *
 */

/*
 * we return an (opaque) pointer to this structure from the
 * VC_OpenProfileAccess function. It contains the information we need
 * to call the registry functions.
 */
struct _VC_PROFILE_INFO {
    DWORD SizeStruct;
    REG_ACCESS RegAccess;
    DWORD dwPlatformId;           
    DEVNODE dnDevNode;            

};

typedef struct _VC_PROFILE_INFO VC_PROFILE_INFO;

#define IsBadVC_PROFILE_INFO(p) ((p) == NULL || IsBadWritePtr((p), sizeof(VC_PROFILE_INFO)) || (p)->SizeStruct != sizeof(VC_PROFILE_INFO))

/*
 * open a handle to whatever functions are needed to access the registry,
 * service controller or profile. Must call this function before
 * calls to the other VC_ configuration routines.
 *
 * For WinNT:
 *  The argument DriverName is the name of the driver. This should be the 
 *  name of the kernel driver file (without path or extension). It will also 
 *  be used as the registry key name or profile section name.
 *
 * For Win95:
 *  The argument dnDevNode indicates a specific plug and play device node 
 *  to supported by the DriverName kernel mode driver. 
 *
 * Even if the open fails, we still return the allocated structure - so
 * that the absence of open handles within the structure can be used
 * by DrvAccess operations to test if we have sufficient privilege.
 */
PVC_PROFILE_INFO WINAPI
VC_OpenProfileAccess(LPTSTR DriverName, DEVNODE dnDevNode)
{
	OSVERSIONINFO VerInfo;
    PVC_PROFILE_INFO pProf;

    /* 
     * We need to get the platform information to determine 
     * if we are running on WinNT or Win95
     */
    VerInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    if (!GetVersionEx(&VerInfo)) {
        return NULL;
    }

    /*
     * allocate the profile info from the global heap
     */
    pProf = GlobalLock(GlobalAlloc(GPTR, sizeof(VC_PROFILE_INFO)));
    if (pProf == NULL) {
        return(NULL);
    }

    /*
     * set the driver name, but don't access the services controller
     * unless we need to (administrators only can do this)
     */
    pProf->SizeStruct = sizeof(VC_PROFILE_INFO);
    lstrcpy (pProf->RegAccess.DriverName, DriverName);
    pProf->RegAccess.ServiceManagerHandle = NULL;
    pProf->dwPlatformId = VerInfo.dwPlatformId;
    pProf->dnDevNode = dnDevNode;

    return(pProf);
}


/*
 * close the handles opened during OpenProfileAccess.
 */
VOID WINAPI
VC_CloseProfileAccess(PVC_PROFILE_INFO pProf)
{
    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return;
    } 

    /* 
     * If the platform is Win95 we let the system (Plug & Play) 
     * handle installing the driver and any basic configuration
     * management (interrupts, port address, memory).
     */ 
    if (pProf->dwPlatformId == VER_PLATFORM_WIN32_NT) {
        /* 
         * close the service controller handle opened 
         * by DrvCreateServiceNode
         */
        DrvCloseServiceManager(&pProf->RegAccess);
    }

    /*
     * free up the profile info structure that we allocated from the global
     * heap.
     */
    pProf->SizeStruct = 0;
    GlobalFree(GlobalHandle(pProf));

    return;
}


/*
 * start the hardware-access portion of the driver. Call the callback
 * function at a moment when it is possible to write configuration information
 * to the profile using VC_WriteProfile.
 * Returns DRVCNF_OK if all is ok, DRVCNF_CANCEL for failure, or DRVCNF_RESTART if
 * all is ok but a system-restart is needed before the driver will load correctly.
 */
LRESULT WINAPI
VC_InstallDriver(
    PVC_PROFILE_INFO pProf,         // access info returned by OpenProfileAccess
    PPROFILE_CALLBACK pCallback,    // callback function
    PVOID pContext                  // context info for callback    
)
{
    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return(VC_ERROR_CREATEDEVICE);
    } 

    /* 
     * If the platform is Win95 we let the system (Plug & Play) 
     * handle installing the driver and any basic configuration
     * management (interrupts, port address, memory).
     */ 
    if (pProf->dwPlatformId != VER_PLATFORM_WIN32_NT) {
        return(VC_OK);
    }

    /*
     * open a handle to the service entry in the registry - force
     * creation of the entry if it does not exist.
     */                                         
    if (!DrvCreateServicesNode(             
            pProf->RegAccess.DriverName, // name of node and driver file    
            SoundDriverTypeNormal,       // determines dependency order
            &pProf->RegAccess,           // open handle info    
            TRUE)) {

        return(VC_INSUFFICIENT_PRIVILEGE);
    }

	/* ensure that the driver is not currently loaded */
    DrvUnloadKernelDriver(&pProf->RegAccess);

    /* now its safe to store items in the registry */
    if (pCallback != NULL) {
        if (!pCallback(pContext)) {
	        return(VC_ERROR_CREATEDEVICE);
        }
    }

    /* now load the driver */
    DrvLoadKernelDriver(&pProf->RegAccess);
	return VC_ReadProfile(pProf, REG_STATUS, VC_ERROR_CREATEDEVICE);
	
}


/*
 * Write a single string keyword and DWORD value to the registry or profile
 * for this driver.
 * This can be re-read from the h/w driver using VC_ReadProfile (in kernel mode).
 *
 * return TRUE for success or FALSE for failure.
 */
BOOL WINAPI
VC_WriteProfile(PVC_PROFILE_INFO pProf, LPTSTR ValueName, DWORD Value)
{
    LONG lRet;

    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return FALSE;
    } 

    /* 
     * If the platform is Win95, this is the same as reading
     * the user profile.
     */ 
    if (pProf->dwPlatformId != VER_PLATFORM_WIN32_NT) {
        return VC_WriteProfileUser(pProf, NULL, ValueName, Value);
    }

    lRet = DrvSetDeviceParameter(&pProf->RegAccess, ValueName, Value);

    if (lRet == ERROR_SUCCESS) {
        return(TRUE);
    } else {
        return(FALSE);
    }
}


/*
 * read back a driver-specific DWORD profile parameter that was written with
 * VC_WriteProfile. If the valuename cannot be found, the default is returned.
 *
 * use RegOpenKeyEx directly instead of via registry.c so that we can get
 * read-only use and permit capture from a non-admin.
 */
DWORD WINAPI
VC_ReadProfile(PVC_PROFILE_INFO pProf, LPTSTR ValueName, DWORD dwDefault)
{
    LONG lRet;
    DWORD dwResult;
    DWORD dwType;
    DWORD dwLength;
    HKEY hkParams;
    TCHAR RegistryPath[MAX_PATH];

    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return dwDefault;
    } 

    /* 
     * If the platform is Win95, this is the same as reading
     * the user profile.
     */ 
    if (pProf->dwPlatformId != VER_PLATFORM_WIN32_NT) {
        return VC_ReadProfileUser(pProf, NULL, ValueName, dwDefault);
    }

    /* Create the path to our node */
    lstrcpy(RegistryPath, TEXT("SYSTEM\\CurrentControlSet\\Services\\"));
    lstrcat(RegistryPath, pProf->RegAccess.DriverName);
    lstrcat(RegistryPath, TEXT("\\"));
    lstrcat(RegistryPath, PARMS_SUBKEY);

    /* See if we can get a registry handle to our device data */
    if (RegOpenKeyEx(
            HKEY_LOCAL_MACHINE,
            RegistryPath,
            0L,
            KEY_QUERY_VALUE,
            &hkParams) != ERROR_SUCCESS) {
        
        return dwDefault;
    }

    dwLength = sizeof(dwResult);  // initialise to size of result field
    lRet = RegQueryValueEx(
            hkParams,
            ValueName,
            NULL,
            &dwType,
            (PCHAR)&dwResult,
            &dwLength);

    RegCloseKey(hkParams);

    if ((lRet != ERROR_SUCCESS) || (dwType != REG_DWORD)) {
        return dwDefault;
    } else {
        return dwResult;
    }
}

static const TCHAR gszUserParmsPath[] = TEXT("Software\\Microsoft\\Multimedia\\Video Capture\\");
#define USERPARMSPATH gszUserParmsPath

/*
 * read back a driver-specific DWORD profile parameter that was written with
 * VC_WriteProfileUser. If the valuename cannot be found, the default is 
 * returned.
 */
DWORD WINAPI
VC_ReadProfileUser(
    PVC_PROFILE_INFO pProf, 
    LPTSTR SectionName, 
    LPTSTR ValueName, 
    DWORD dwDefault
)
{
    LONG lRet;
    DWORD dwResult;
    DWORD dwType;
    DWORD dwLength;
    HKEY hkParams, hkSystem;
    TCHAR RegistryPath[MAX_PATH];

    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return dwDefault;
    } 

    /* Create the path to our node */
    lstrcpy(RegistryPath, USERPARMSPATH);
    lstrcat(RegistryPath, pProf->RegAccess.DriverName);

   /* 
    * if no SectionName provided - use the base
    * key
    */
	if (SectionName != NULL) {
		lstrcat(RegistryPath, "\\");
        lstrcat(RegistryPath, SectionName);
    }

    hkSystem = HKEY_CURRENT_USER;

    /* See if we can get a registry handle to our device data */
    if (RegOpenKeyEx(
            hkSystem,
            RegistryPath,
            0L,
            KEY_QUERY_VALUE,
            &hkParams) != ERROR_SUCCESS) {
        return dwDefault;
    }

    dwLength = sizeof(dwResult);  // initialise to size of result field
    lRet = RegQueryValueEx(
            hkParams,
            ValueName,
            NULL,
            &dwType,
            (PCHAR)&dwResult,
            &dwLength);

    RegCloseKey(hkParams);

    if ((lRet != ERROR_SUCCESS) || (dwType != REG_DWORD)) {
        return dwDefault;
    } else {
        return dwResult;
    }
}


/*
 * write a driver-specific DWORD profile parameter for current user use.
 */
BOOL WINAPI
VC_WriteProfileUser(
    PVC_PROFILE_INFO pProf, 
    LPTSTR SectionName, 
    LPTSTR ValueName, 
    DWORD Value
)
{
    LONG lRet;
    DWORD dwDisposition;
    HKEY hkParams, hkSystem;
    TCHAR RegistryPath[MAX_PATH];

    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return FALSE;
    } 

    /* Create the path to our node */
    lstrcpy(RegistryPath, USERPARMSPATH);
    lstrcat(RegistryPath, pProf->RegAccess.DriverName);

    /* 
     * if no SectionName provided - just use the base
     * key
     */
    if (SectionName != NULL) {
        lstrcat(RegistryPath, "\\");
        lstrcat(RegistryPath, SectionName);
    }

    hkSystem = HKEY_CURRENT_USER;

    /* See if we can get a registry handle to our device data */
    if (ERROR_SUCCESS == RegCreateKeyEx(
            hkSystem,
            RegistryPath,
            0L,
            NULL,       // no class name
            0,          // no special options (this is NOT volatile data)
            KEY_WRITE,
            NULL,       // no special security
            &hkParams,
            &dwDisposition)) {

        lRet = RegSetValueEx(
                hkParams,
                ValueName,
                0,
                REG_DWORD,
                (PCHAR)&Value,
                sizeof(Value));

        RegCloseKey(hkParams);

        if (ERROR_SUCCESS == lRet) {
            return TRUE;
        }
    }

    return(FALSE);
}


/*
 * read a string parameter from the device's profile. returns FALSE
 * if it fails to read the string.
 */
BOOL WINAPI
VC_ReadProfileString(
    PVC_PROFILE_INFO pProf,
    LPTSTR ValueName,
    LPTSTR ValueString,
    DWORD ValueLength
)
{
    LONG lRet;
    DWORD dwType;
    HKEY hkParams, hkSystem;
    TCHAR RegistryPath[MAX_PATH];

    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return FALSE;
    } 

    /*
     * fill the whole string with 0s to ensure that it is null
     * terminated on return
     */
    FillMemory(ValueString, ValueLength, 0);

    /* Create the path to our node */
    lstrcpy(RegistryPath, TEXT("SYSTEM\\CurrentControlSet\\Services\\"));
	lstrcat(RegistryPath, pProf->RegAccess.DriverName);
	lstrcat(RegistryPath, TEXT("\\"));
	lstrcat(RegistryPath, PARMS_SUBKEY);
    hkSystem = HKEY_LOCAL_MACHINE;

    /* See if we can get a registry handle to our device data */
    if (RegOpenKeyEx(
            hkSystem,
            RegistryPath,
            0L,
            KEY_QUERY_VALUE,
            &hkParams) != ERROR_SUCCESS) {

        return FALSE;
    }

    lRet = RegQueryValueEx(
            hkParams,
            ValueName,
            NULL,
            &dwType,
            (PUCHAR) ValueString,
            &ValueLength);

        RegCloseKey(hkParams);


    if ((lRet != ERROR_SUCCESS) || (dwType != REG_SZ)) {
        return(FALSE);
    } else {
        return(TRUE);
    }
}


/*
 * read back a driver-specific string profile parameter that was written with
 * VC_WriteProfileUserString. If the valuename cannot be found, the default 
 * is returned.
 */
BOOL WINAPI
VC_ReadProfileUserString(
    PVC_PROFILE_INFO pProf,
    LPTSTR SectionName, 
    LPTSTR ValueName,
    LPTSTR ValueString,
    DWORD ValueLength
)
{
    LONG lRet;
    DWORD dwType;
    HKEY hkParams, hkSystem;
    TCHAR RegistryPath[MAX_PATH];

    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return FALSE;
    } 

    /*
     * fill the whole string with 0s to ensure that it is null
     * terminated on return
     */
    FillMemory(ValueString, ValueLength, 0);

    /* Create the path to our node */
    lstrcpy(RegistryPath, USERPARMSPATH);
    lstrcat(RegistryPath, pProf->RegAccess.DriverName);

    /* 
     * if no SectionName provided - use the base
     * key
     */
    if (SectionName != NULL) {
        lstrcat(RegistryPath, "\\");
        lstrcat(RegistryPath, SectionName);
    }

    hkSystem = HKEY_CURRENT_USER;

    /* See if we can get a registry handle to our device data */
    if (RegOpenKeyEx(
            hkSystem,
            RegistryPath,
            0L,
            KEY_QUERY_VALUE,
            &hkParams) != ERROR_SUCCESS) {
        return FALSE;
    }

    lRet = RegQueryValueEx(
            hkParams,
            ValueName,
            NULL,
            &dwType,
            ValueString,
            &ValueLength);

        RegCloseKey(hkParams);


    if ((lRet != ERROR_SUCCESS) || (dwType != REG_SZ)) {
	    FillMemory(ValueString, ValueLength, 0);
        return(FALSE);
    } else {
        return(TRUE);
    }
}


/*
 * write a driver-specific string profile parameter for current user use.
 */
BOOL WINAPI
VC_WriteProfileUserString(
    PVC_PROFILE_INFO pProf, 
    LPTSTR SectionName, 
    LPTSTR ValueName, 
    LPTSTR ValueString,
    DWORD ValueLength
)
{
    LONG lRet;
    DWORD dwDisposition;
    HKEY hkParams, hkSystem;
    TCHAR RegistryPath[MAX_PATH];

    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return FALSE;
    } 

    /* Create the path to our node */
    lstrcpy(RegistryPath, USERPARMSPATH);
    lstrcat(RegistryPath, pProf->RegAccess.DriverName);

    /* 
     * if no SectionName provided - just use the base
     * key
     */
    if (SectionName != NULL) {
        lstrcat(RegistryPath, "\\");
        lstrcat(RegistryPath, SectionName);
    }

    hkSystem = HKEY_CURRENT_USER;

    /* See if we can get a registry handle to our device data */
    if (ERROR_SUCCESS == RegCreateKeyEx(
            hkSystem,
            RegistryPath,
            0L,
            NULL,       // no class name
            0,          // no special options (this is NOT volatile data)
            KEY_WRITE,
            NULL,       // no special security
            &hkParams,
            &dwDisposition)) {

        lRet = RegSetValueEx(
                hkParams,
                ValueName,
                0,
                REG_SZ,
                ValueString,
                ValueLength);

        RegCloseKey(hkParams);

        if (ERROR_SUCCESS == lRet) {
            return TRUE;
        }
    }

    return(FALSE);
}


/*
 * unload a driver. On NT, this stops and removes the kernel-mode driver.
 * On win95 this funcion does nothing.
 *
 * the VC_PROFILE_INFO structure is not valid after this call.
 *
 * return DRVCNF_OK if the unload was successful, DRVCNF_CANCEL if it failed, and
 * DRVCNF_RESTART if a system-restart is needed before the removal takes effect.
 */
LRESULT WINAPI
VC_RemoveDriver(PVC_PROFILE_INFO pProf)
{
    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return DRVCNF_CANCEL;
    } 

    /* 
     * If the platform is Win95 we let the system (Plug & Play) 
     * handle installing the driver and any basic configuration
     * management (interrupts, port address, memory).
     */ 
    if (pProf->dwPlatformId != VER_PLATFORM_WIN32_NT) {
        return(DRVCNF_OK);
    }

    return(DrvRemoveDriver(&pProf->RegAccess));
}


/*
 * do we have sufficient privilege to access the services controller,
 * registry and anything else we need to configure the driver ?
 */
BOOL WINAPI
VC_ConfigAccess(PVC_PROFILE_INFO pProf)
{
    /* verify profile handle */
    if (IsBadVC_PROFILE_INFO(pProf)) {
        return FALSE;
    } 

    /* Access to the registry is not secure in Win95 */
    if (pProf->dwPlatformId != VER_PLATFORM_WIN32_NT) {
        return TRUE;    
    }

    /*
     * open the services node without creating it, to init
     * the RegAccess structure
     */
    DrvCreateServicesNode(
            pProf->RegAccess.DriverName, // file/node name
            SoundDriverTypeNormal,       // decides starting order/dependency
            &pProf->RegAccess,           // storage for registry handles etc
            FALSE);                      // don't create node if not there

    return(DrvAccess(&pProf->RegAccess));
}


/*
 * Entry point to hardware specific function for displaying any 
 * required driver installation dialog box.
 */
LRESULT WINAPI 
VC_DriverInstallDialog (HWND hWnd, LPCONFIG_INFO lpGeneric)
{
	/* call the hardware specific dialog function */
	return DriverInstallDialog (hWnd, lpGeneric);

}


/*
 * Entry point to hardware specific function for displaying any 
 * required driver properties dialog box.
 */
LRESULT WINAPI
VC_DriverPropertiesDialog (HWND hWnd, LPCONFIG_INFO lpGeneric)
{
	/* call the hardware specific dialog function */
	return DriverPropertiesDialog (hWnd, lpGeneric);
}

