/*
 * vcworker.c
 *
 * 32-bit Video Capture driver
 * User-mode support library - worker functions
 *
 *
 * Mark Vrabel, October 95:
 */

#include <vcdriver.h>
#include "vcupriv.h"


BOOL VC_ServiceInterrupt (VCUSER_HANDLE vh)
{
    BOOL bRepeat, bServiced;

    /*
     * call the acknowledge. this will not do any service, but will
     * return TRUE if the service routine is to be called.
     */
    if (vh->Callback.InterruptAcknowledge == NULL) {

        /* no isr - and yet we are getting interrupts -
         * we should disconnect the interrupt
         */

        return FALSE;
    }

    /*
     * call the acknowledge. this will not do any service, but will
     * return TRUE if the service routine is to be called.
     */
    while (vh->Callback.InterruptAcknowledge(vh, &bServiced, &bRepeat)) {
        
		LPVIDEOHDR lpVideoHdr;
		BOOL bKeyFrame = FALSE;

		BOOL bUseCallback = (HIWORD(vh->UserCallback.dwFlags != 0) || vh->UserCallback.dwCallback != 0) ?TRUE:FALSE;

        /* 
         * make sure we have a capture service installed 
         */
        if (vh->Callback.CaptureService == NULL) {
            return FALSE;
        }
        
		/*
	     * get the next buffer from the queue
		 */
        lpVideoHdr = VC_RemoveReadyHead(vh);
		if (lpVideoHdr) {

	        /*
		     * find the length of the data buffer
			 */
			lpVideoHdr->dwBytesUsed = (DWORD) vh->Callback.CaptureService(
						vh,
                        (LPUCHAR)lpVideoHdr->dwReserved[HEADER_DATA],
                        (LPULONG) &lpVideoHdr->dwTimeCaptured,
						lpVideoHdr->dwBufferLength,
						&bKeyFrame);
        
            /* if a capture service routine returns 0, it means
	         * that it has not completed this capture, and would like
		     * the same buffer back for the next field if possible
			 * - in this case, return the buffer to the start of the
			 * queue
			 */
			if (lpVideoHdr->dwBytesUsed == 0) {
				VC_ReplaceReadyHead(vh, lpVideoHdr);
				continue;
			} else {
			
				/* if it's a key frame - mark it */
				if (bKeyFrame) {
				    lpVideoHdr->dwFlags |= VHDR_KEYFRAME;
				}
			}

			/* post the first buffer as done */
			if (bUseCallback) {
				VC_AddDoneTail(vh, lpVideoHdr);
				ReleaseSemaphore(vh->hBufferSemaphore, 1, NULL);
			} else {
			    lpVideoHdr->dwFlags |= VHDR_DONE;
				lpVideoHdr->dwFlags &= ~VHDR_INQUEUE;
			}
		} else {

			/* report the skipped frame */
		    if (vh->Callback.StreamReportSkipFunc != NULL) {
				vh->Callback.StreamReportSkipFunc (vh, 1);
			}
			
			if (bUseCallback) {
				ReleaseSemaphore(vh->hErrorSemaphore, 1, NULL);
			} else {
			    vh->SkipCount++;
			}

			/* no need to continue */
			return bServiced;
		}

		/* do we need to repeat the last frame??? */
		if (bRepeat && vh->Callback.RepeatFrame != NULL) {

			BOOL bRepeatComplete = FALSE;

			/* 
			 * we may need to repeat this frame more than once to maintain 
			 * audio video synch
			 */
			do {

				/* audio/ vidio syncronization frame repeat required */
				LPVIDEOHDR lpRepeatHdr = VC_RemoveReadyHead(vh);
				if (lpRepeatHdr) {

					/*
					 * try to repeat the previous frame
					 */
					lpRepeatHdr->dwBytesUsed = 
					(DWORD) vh->Callback.RepeatFrame(
								vh,
								(LPUCHAR)lpVideoHdr->dwReserved[HEADER_DATA],
								lpVideoHdr->dwBytesUsed,
								(LPUCHAR)lpRepeatHdr->dwReserved[HEADER_DATA],
								lpRepeatHdr->dwBufferLength,
								(LPULONG) &lpRepeatHdr->dwTimeCaptured);

					if (lpRepeatHdr->dwBytesUsed == 0) {
						bRepeatComplete = TRUE;
						VC_ReplaceReadyHead(vh, lpRepeatHdr);
					} else {

						/* if it's a key frame - mark it */
						if (bKeyFrame) {
							lpRepeatHdr->dwFlags |= VHDR_KEYFRAME;
						}

						/* post the second buffer as done */
						if (bUseCallback) {
							VC_AddDoneTail(vh, lpRepeatHdr);
							ReleaseSemaphore(vh->hBufferSemaphore, 1, NULL);
						} else {
							lpRepeatHdr->dwFlags &= ~VHDR_INQUEUE;
							lpRepeatHdr->dwFlags |= VHDR_DONE;
						}
					}

				} else {

					/* report the skipped frame */
				    if (vh->Callback.StreamReportSkipFunc != NULL) {
						vh->Callback.StreamReportSkipFunc (vh, 1);
					}

					if (bUseCallback) {
						ReleaseSemaphore(vh->hErrorSemaphore, 1, NULL);
					} else {
						vh->SkipCount++;
					}

					bRepeatComplete = TRUE;
				}

			} while (!bRepeatComplete);
		} 
    }

	return bServiced;
}


/* 
 *
 */
VOID CALLBACK
VC_TimerProc(
    UINT  wTimerId,    
    UINT  wMsg,
    DWORD dwUser,
    DWORD dw1,
    DWORD dw2         
)
{
    VCUSER_HANDLE vh = (VCUSER_HANDLE)dwUser;
	BOOL bServiced;

    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return;
    }

    /* make sure we should be running */
    if (vh->TimerId != wTimerId) {
		SetEvent(vh->hTimerEvent);
        return;
    }

    /* is the timer isr disabled??? */
    if (vh->bTimerPaused) {
        /* restart the timer */
        vh->TimerId = timeSetEvent (vh->Delay, vh->Resolution, VC_TimerProc, (DWORD)vh, TIME_ONESHOT);
    	return;
	}

	/* service the interrupt */
	bServiced = VC_ServiceInterrupt (vh);

    /* if we we're not serviced we shouldn't restart!!! */
    if (bServiced && vh->TimerId != 0) {

        /* restart the timer */
        vh->TimerId = timeSetEvent (vh->Delay, vh->Resolution, VC_TimerProc, (DWORD)vh, TIME_ONESHOT);
    }

}



#define VC_TERMINATE    (WAIT_OBJECT_0) 
#define VC_ERROR        (WAIT_OBJECT_0+1)
#define VC_BUFFER_DONE  (WAIT_OBJECT_0+2)


DWORD
VC_ThreadInit(DWORD arg)
{
    VCUSER_HANDLE vh = (VCUSER_HANDLE) arg;
    LPVIDEOHDR lpVideoHdr;
    BOOL bTerminate = FALSE;
    DWORD hWait;
    
    vh->FunctionResult = TRUE;
    if (vh->Callback.StreamInitFunc != NULL) {
        vh->FunctionResult =  vh->Callback.StreamInitFunc(vh, vh->FunctionArg);
        if (!vh->FunctionResult) {
            bTerminate = TRUE;
        } 
    }
    
    /* create some semaphores for tracking buffers and errors */
    vh->hBufferSemaphore = CreateSemaphore(NULL, 0, 0x7fffffff, NULL);
    if (vh->hBufferSemaphore == NULL) {
        bTerminate = TRUE;
    } else {
        vh->hErrorSemaphore = CreateSemaphore(NULL, 0, 0x7fffffff, NULL);
        if (vh->hErrorSemaphore == NULL) {
            bTerminate = TRUE;
        }
    }

    /* are we ready to begin??? */
    if (!bTerminate) {
        /* create a multiple wait object */
        vh->hWaitObjects[0] = vh->hWorkEvent;
        vh->hWaitObjects[1] = vh->hErrorSemaphore;
        vh->hWaitObjects[2] = vh->hBufferSemaphore;

        vh->State = State_Init;
        vh->SkipCount = 0;
    }

    /* signal that init is done */
    VC_Callback(&vh->UserCallback, MM_DRVM_OPEN, 0);
    SetEvent(vh->hCompleteEvent);

    while (!bTerminate) {

        /* wait for completion of the operation */
        hWait = WaitForMultipleObjects(3, vh->hWaitObjects, FALSE, INFINITE);
        
        switch (hWait) { 
            
            case VC_TERMINATE:
                bTerminate = TRUE;
                break;

            case VC_ERROR:
                if (vh->State > State_Idle) {
                    VC_SkipCount(vh, 1);
                }
                break;

            case VC_BUFFER_DONE:
                
                if (vh->State == State_Start) {
                
                    lpVideoHdr = VC_RemoveDoneHead(vh);

                    /* 
                     * if we've been signaled, we should have
                     * a buffer - it should never return NULL!!!
                     */ 
                    if (lpVideoHdr != NULL) {
                        VC_ProcessBuffer(vh, lpVideoHdr);
                    }
                }
                break;
        }       
	}


    if (vh->hErrorSemaphore != NULL) {
        CloseHandle(vh->hErrorSemaphore);
    }
    if (vh->hBufferSemaphore != NULL) {
        CloseHandle(vh->hBufferSemaphore);
    }

    VC_Callback(&vh->UserCallback, MM_DRVM_CLOSE, 0);

    return(0);
}


/*
 * routines to start and stop the event timer
 */
BOOL
VC_StartEventTimer (VCUSER_HANDLE vh, UINT Delay)
{
    /* we're already running - don't start another one!!! */
    if (vh->TimerId != 0) {
        return FALSE;
    }
	
	/* start 'er up... */
    vh->Delay = Delay;
	vh->TimerId = timeSetEvent (vh->Delay, vh->Resolution, VC_TimerProc, (DWORD)vh, TIME_ONESHOT);
                
    if (vh->TimerId == 0) {
        return FALSE;
    }

    return TRUE;
}


VOID
VC_StopEventTimer (VCUSER_HANDLE vh)
{
    UINT TimerId = vh->TimerId;

	vh->TimerId = 0;
	WaitForSingleObject (vh->hTimerEvent, vh->Delay*100);    
	
}

/*
 * Disable event timer from responding to interrupt
 * (Psuedo interrupt disable) 
 */
void 
VC_TimerDisable (VCUSER_HANDLE vh)
{
    vh->bTimerPaused = TRUE;
}


/*
 * enable event timer to respond to interrupt
 * (Psuedo interrupt enable) 
 */
void 
VC_TimerEnable (VCUSER_HANDLE vh)
{
	vh->bTimerPaused = FALSE;
}


/*
 * send a callback to the user.
 * if callback is null or flags
 * are 0, no callback required.
 */
VOID 
VC_Callback(LPVCCALLBACK lpCallback, UINT msg, DWORD data)
{
    if (lpCallback == NULL) {
        return;
    }
	
	if (HIWORD(lpCallback->dwFlags == 0) || lpCallback->dwCallback == 0) {
		return;
	}
	
    DriverCallback(lpCallback->dwCallback,
            HIWORD(lpCallback->dwFlags),
            (HANDLE)lpCallback->hDevice,
            msg,
            lpCallback->dwUser,
            data,
            0            // not used
            );
}


/*
 * one of the buffers has completed.
 *
 * callback if necessary
 */
VOID
VC_ProcessBuffer(VCUSER_HANDLE vh, LPVIDEOHDR lpVideoHdr)
{
    /* mark buffer done */
    lpVideoHdr->dwFlags |= VHDR_DONE;

    /* callback if necessary */
    VC_Callback(&vh->UserCallback, MM_DRVM_DATA, (DWORD) lpVideoHdr);
}


/*
 * we have got a skip-count back from the driver. Increment our
 * local count and callback if necessary
 */
VOID
VC_SkipCount(VCUSER_HANDLE vh, DWORD dwSkips)
{
    vh->SkipCount += dwSkips;

    if (dwSkips > 0) {
        VC_Callback(&vh->UserCallback, MM_DRVM_ERROR, vh->SkipCount);
    }
}


/*
 * --  buffer-queueing functions -----------------------
 *
 * maintain a single-linked list of LPVIDEOHDR buffer headers for which
 * we can extract the first one or add to the end.
 *
 * we use the dwReserved[0] field as the link. vh->ReadyListHead points to the
 * first item in the list. vh->ReadyListTail points to the last item in the list
 * which itself points to NULL. for an empty list, both ListHead and ListTail
 * are null.
 */

#define NEXT_LPVH(p)    (LPVIDEOHDR)((p)->dwReserved[0])

/*
 * extract the first item in the list of LPVIDEOHDRs. NULL if none in list
 */
LPVIDEOHDR
VC_RemoveReadyHead(VCUSER_HANDLE vh)
{
	LPVIDEOHDR lpvh;

    if (vh->ReadyListHead == NULL) {
        return(NULL);
    }

    lpvh = vh->ReadyListHead;
    vh->ReadyListHead = NEXT_LPVH(lpvh);

    if (vh->ReadyListHead == NULL) {
        vh->ReadyListTail = NULL;
    }

    if (vh->Callback.StreamRemoveReadyHeadFunc != NULL) {
		vh->Callback.StreamRemoveReadyHeadFunc (vh, lpvh);
    }

    return(lpvh);
}


/*
 * add an item to the tail of the list
 */
VOID
VC_AddReadyTail(VCUSER_HANDLE vh, LPVIDEOHDR lpvh)
{
    if (vh->Callback.StreamAddReadyTailFunc != NULL) {
		vh->Callback.StreamAddReadyTailFunc (vh, lpvh);
    }

	lpvh->dwFlags &= ~VHDR_KEYFRAME;
    lpvh->dwFlags &= ~VHDR_DONE;
    lpvh->dwFlags |= VHDR_INQUEUE;

    lpvh->dwBytesUsed = 0;
    
    if (vh->ReadyListHead == NULL) {
        vh->ReadyListHead = lpvh;
    } else {
        NEXT_LPVH(vh->ReadyListTail) = lpvh;
    }

    NEXT_LPVH(lpvh) = NULL;
    vh->ReadyListTail = lpvh;

}


/*
 * replace an item at the head of the list
 */
VOID
VC_ReplaceReadyHead(VCUSER_HANDLE vh, LPVIDEOHDR lpvh)
{
    if (vh->Callback.StreamReplaceReadyHeadFunc != NULL) {
		vh->Callback.StreamReplaceReadyHeadFunc (vh, lpvh);
    }

    if (vh->ReadyListHead == NULL) {
        VC_AddReadyTail(vh, lpvh);
    } else {
        NEXT_LPVH(lpvh) = vh->ReadyListHead;
		vh->ReadyListHead = lpvh;
    }
}


LPVIDEOHDR
VC_RemoveDoneHead(VCUSER_HANDLE vh)
{
    LPVIDEOHDR lpvh;

    if (vh->DoneListHead == NULL) {
        return(NULL);
    }

    lpvh = vh->DoneListHead;
    vh->DoneListHead = NEXT_LPVH(lpvh);

    if (vh->DoneListHead == NULL) {
        vh->DoneListTail = NULL;
    }
    
    lpvh->dwFlags &= ~VHDR_INQUEUE;

    return(lpvh);
}


VOID
VC_AddDoneTail(VCUSER_HANDLE vh, LPVIDEOHDR lpvh)
{
    if (vh->DoneListHead == NULL) {
        vh->DoneListHead = lpvh;
    } else {
        NEXT_LPVH(vh->DoneListTail) = lpvh;
    }

    NEXT_LPVH(lpvh) = NULL;
    vh->DoneListTail = lpvh;
}


VOID
VC_ReplaceDoneHead(VCUSER_HANDLE vh, LPVIDEOHDR lpvh)
{
    if (vh->DoneListHead == NULL) {
        VC_AddDoneTail(vh, lpvh);
    } else {
        NEXT_LPVH(lpvh) = vh->DoneListHead;
    vh->DoneListHead = lpvh;
    }
}


/*
 * clear out our local queue of add-buffer requests that
 * have not yet been sent to the driver.
 *
 */
VOID VC_ResetReadyQueue(VCUSER_HANDLE vh)
{
    LPVIDEOHDR lpvh;

    while ( (lpvh = VC_RemoveReadyHead(vh)) != NULL) {

        /* mark buffer done */
        lpvh->dwFlags |= VHDR_DONE;
        lpvh->dwFlags &= ~VHDR_INQUEUE;

        /* callback if necessary */
        VC_Callback(&vh->UserCallback, MM_DRVM_DATA, (DWORD) lpvh);
    }
}


VOID VC_ResetDoneQueue(VCUSER_HANDLE vh)
{
    LPVIDEOHDR lpvh;

    while ( (lpvh = VC_RemoveDoneHead(vh)) != NULL) {

        /* mark buffer done */
        lpvh->dwFlags |= VHDR_DONE;
        lpvh->dwFlags &= ~VHDR_INQUEUE;

        /* callback if necessary */
        VC_Callback(&vh->UserCallback, MM_DRVM_DATA, (DWORD) lpvh);
    }
}


